from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Callable, Optional

import numpy as np

from opencosmo.index import SimpleIndex
from opencosmo.io.updaters import apply_updaters

if TYPE_CHECKING:
    import h5py

    from opencosmo.header import OpenCosmoHeader
    from opencosmo.index import DataIndex
    from opencosmo.io import protocols as iop

try:
    from mpi4py import MPI
except ImportError:
    MPI = None  # type: ignore

"""
Writers work in tandem with schemas to create new files. All schemas must have
an into_writer method, which returns a writer that can be used to put
data into the new file.

Schemas are responsible for validating and building the file structure 
as well as allocating space.  As a result, writers ASSUME the correct structure exists, 
and that all the datasets have the correct size, datatype, etc.
"""


def write_index(
    input_ds: h5py.Dataset | np.ndarray | None,
    output_ds: h5py.Dataset,
    index: DataIndex,
    offset: int = 0,
    updater: Optional[Callable[[np.ndarray], np.ndarray]] = None,
):
    """
    Helper function to take elements from one h5py.Dataset using an index
    and put it in a different one.
    """
    data = np.array([])
    if len(index) > 0 and input_ds is not None:
        data = index.get_data(input_ds)
        if updater is not None:
            data = updater(data)

        data = data.astype(input_ds.dtype)

    if input_ds is not None:
        output_ds[offset : offset + len(data)] = data


class FileWriter:
    """
    Root writer for a file. Pretty much just calls the child writers.
    """

    def __init__(self, children: dict[str, iop.DataWriter]):
        self.children = children

    def write(self, file: h5py.File):
        if len(self.children) == 1:
            ds = next(iter(self.children.values()))
            return ds.write(file)
        for name, dataset in self.children.items():
            dataset.write(file[name])


class CollectionWriter:
    """
    Writes collections to a file or grous. Also pretty much just calls
    the child writers. May or may not recieve a header to write, depending
    on they type of collection.
    """

    def __init__(
        self,
        children: dict[str, iop.DataWriter],
        header: Optional[OpenCosmoHeader] = None,
    ):
        self.children = children
        self.header = header

    def write(self, file: h5py.File | h5py.Group):
        if len(self.children) == 1:
            next(iter(self.children.values())).write(file)
            return

        child_names = list(self.children.keys())
        child_names.sort()
        for name in child_names:
            self.children[name].write(file[name])


class DatasetWriter:
    """
    Writes datasets to a file or group. Datasets must have at least one column.
    If the datset is being written alone or as part of SimulationCollection, it will
    be responsible for writing a header.

    It may or may not have a spatial index. It also may or may not have links
    to other datasets.
    """

    def __init__(
        self,
        columns: dict[str, dict[str, ColumnWriter]],
        comm: Optional[MPI.Comm] = None,
    ):
        self.columns = apply_updaters(columns, comm)
        self.comm = comm

    def write(self, group: h5py.Group):
        groupnames = list(self.columns.keys())
        groupnames.sort()
        for groupname in groupnames:
            colnames = list(self.columns[groupname].keys())
            colnames.sort()
            data_group = group[groupname]
            for colname in colnames:
                self.columns[groupname][colname].write(data_group)


class EmptyColumnWriter:
    def __init__(self, name: str):
        self.name = name

    def write(
        self,
        group: h5py.Group,
        updater: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ):
        ds = group[self.name]

        write_index(None, ds, SimpleIndex.empty(), updater=updater)


class ColumnWriter:
    """
    Writes a single column in a dataset. This is the only writer that actually moves
    real data around.
    """

    def __init__(
        self,
        name: str,
        index: DataIndex,
        source: h5py.Dataset,
        offset: int = 0,
        updater: Optional[Callable] = None,
    ):
        self.name = name
        self.source = source
        self.index = index
        self.offset = offset
        self.updater = updater
        self.data = None

    def write(
        self,
        group: h5py.Group,
    ):
        ds = group[self.name]

        write_index(self.source, ds, self.index, self.offset, self.updater)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Mapping, Optional

from opencosmo.units import UnitConvention
from opencosmo.units.get import UnitApplicator, get_unit_applicators_hdf5

if TYPE_CHECKING:
    import astropy.units as u
    import h5py
    import numpy as np
    from astropy.cosmology import Cosmology

    from opencosmo.header import OpenCosmoHeader


def make_unit_handler(
    group: h5py.Group,
    header: "OpenCosmoHeader",
    target_convention: Optional[UnitConvention] = None,
):
    applicators = get_unit_applicators_hdf5(group, header)
    if target_convention is None:
        target_convention = header.file.unit_convention
    return UnitHandler(
        header.file.unit_convention, target_convention, header.cosmology, applicators
    )


class UnitHandler:
    def __init__(
        self,
        base_convention: UnitConvention,
        current_convention: UnitConvention,
        cosmology: Cosmology,
        applicators: dict[str, UnitApplicator],
        conversions: dict[str, u.Unit] = {},
        column_conversions: dict[str, u.Unit] = {},
    ):
        self.__base_convention = base_convention
        self.__current_convention = current_convention
        self.__cosmology = cosmology
        self.__applicators = applicators
        self.__conversions = conversions
        self.__column_conversions = column_conversions

    @property
    def current_convention(self):
        return self.__current_convention

    @cached_property
    def base_units(self):
        return {key: app.base_unit for key, app in self.__applicators.items()}

    @cached_property
    def unitless_columns(self):
        return set(
            key for key, app in self.__applicators.items() if app.base_unit is None
        )

    def with_static_columns(self, **columns: u.Unit):
        new_applicators = {}
        for colname, unit in columns.items():
            new_applicators[colname] = UnitApplicator.static(
                unit, self.__base_convention
            )
        return UnitHandler(
            self.__base_convention,
            self.__current_convention,
            self.__cosmology,
            self.__applicators | new_applicators,
            self.__conversions,
            self.__column_conversions,
        )

    def with_new_columns(self, **columns: u.Unit):
        new_applicators = {}
        for colname, unit in columns.items():
            new_applicators[colname] = UnitApplicator.from_unit(
                unit, self.__base_convention, self.__cosmology
            )
        return UnitHandler(
            self.__base_convention,
            self.__current_convention,
            self.__cosmology,
            self.__applicators | new_applicators,
            self.__conversions,
            self.__column_conversions,
        )

    def verify_conversions(
        self, conversions: dict[str, u.Unit], convention: UnitConvention
    ):
        conversion_keys = set(conversions.keys())
        unitless_columns = self.unitless_columns.intersection(conversion_keys)
        if unitless_columns:
            raise ValueError(f"Columns {unitless_columns} do not carry units!")

        for name, unit in conversions.items():
            if not self.__applicators[name].can_convert(unit, convention):
                raise ValueError(
                    f"Cannot convert values with units {self.__applicators[name].unit_in_convention(convention)} to value with units {unit}"
                )

    def with_convention(self, convention: str | UnitConvention):
        convention = UnitConvention(convention)
        return UnitHandler(
            self.__base_convention, convention, self.__cosmology, self.__applicators
        )

    def with_conversions(
        self, conversions: dict[u.Unit, u.Unit], columns: dict[str, u.Unit]
    ):
        if not conversions and not columns:
            return self

        self.verify_conversions(columns, self.__current_convention)
        new_column_conversions = self.__column_conversions | columns
        new_conversions = self.__conversions | {
            str(key): value for key, value in conversions.items()
        }
        return UnitHandler(
            self.__base_convention,
            self.__current_convention,
            self.__cosmology,
            self.__applicators,
            new_conversions,
            new_column_conversions,
        )

    def into_base_convention(
        self,
        data: Mapping[str, float | np.ndarray | u.Quantity],
        unit_kwargs: dict = {},
    ):
        return {
            name: self.__applicators[name].convert_to_base(
                val, self.__current_convention, unit_kwargs=unit_kwargs
            )
            for name, val in data.items()
        }

    def apply_units(self, data: dict[str, np.ndarray], unit_kwargs):
        if self.__current_convention == UnitConvention.UNITLESS:
            return data
        columns = {}
        for colname, value in data.items():
            applicator = self.__applicators.get(colname)
            column_conversion = self.__column_conversions.get(colname)
            if applicator is not None and applicator.base_unit is not None:
                unitful_value = applicator.apply(
                    value, self.__current_convention, unit_kwargs=unit_kwargs
                )
                unit_conversion = self.__conversions.get(str(unitful_value.unit))
                if unit_conversion is not None and column_conversion is None:
                    unitful_value = unitful_value.to(unit_conversion)
                elif column_conversion is not None:
                    unitful_value = unitful_value.to(column_conversion)
                columns[colname] = unitful_value
            else:
                columns[colname] = value

        return columns



import h5py

from opencosmo.analysis.install import get_specs


def get_file_versions(spec_name, file):
    file = h5py.File(file)
    specs = get_specs()
    spec = specs[spec_name]
    if spec.header_version_key is None:
        raise ValueError("This spec does not support file versions")
    if "header" not in file.keys():
        group = next(iter(file.values()))
    else:
        group = file

    if "header" not in group.keys():
        raise NotImplementedError

    versions = dict(group["header"][spec.header_version_key].attrs)
    file.close()

    return versions



from __future__ import annotations

from functools import singledispatchmethod
from typing import TYPE_CHECKING, Any, Iterable, TypeVar

import astropy.units as u  # type: ignore
import numpy as np
from astropy.coordinates import SkyCoord  # type: ignore
from healpy import ang2vec, query_disc  # type: ignore

from opencosmo.spatial.models import (
    BoxRegionModel,
    ConeRegionModel,
    HealPixRegionModel,
)

if TYPE_CHECKING:
    from astropy.cosmology import FLRW
    from numpy.typing import NDArray

    from opencosmo.spatial.protocols import Region
    from opencosmo.units import UnitConvention
    from opencosmo.units.handler import UnitHandler

T = TypeVar("T", float, u.Quantity)

Point3d = tuple[float, float, float]
Point2d = tuple[T, T]
BoxSize = tuple[float, float, float]


def physical_to_scalefree(value: float, cosmology: FLRW, z: float):
    a = cosmology.scale_factor(z)
    comoving_value = value / a
    return comoving_to_scalefree(comoving_value, cosmology)


def comoving_to_scalefree(value: float, cosmology: FLRW):
    h = cosmology.h
    scalefree_value = value * h
    return scalefree_value


class ConeRegion:
    """
    Cone region for querying lightcones. Defined by RA/Dec coordinate and an angular
    size. Should always be constructed with :py:meth:`opencosmo.make_cone`

    Note that the underlying coordinate representation of the data may or may not be
    in RA and Dec. Spatial queries handle all the necessary conversions, but this may
    mean that the coordinates in the query output do not appear to match the coordinates
    of the region used to perform the query.
    """

    def __init__(self, center: SkyCoord, radius: u.Quantity):
        self.__center = center
        self.__radius = radius

    def __hash__(self):
        return hash(
            (self.__center.ra.deg, self.center.dec, self.radius.to(u.deg).value)
        )

    def into_base_convention(self, *args, **kwargs):
        return self

    def into_model(self) -> ConeRegionModel:
        return ConeRegionModel(
            center=(self.center.ra.value, self.center.dec.value),
            radius=self.__radius.value,
        )

    @property
    def center(self):
        """
        The center of this region.

        Returns
        -------
        coordinate: astropy.coordinates.SkyCoord
        """
        return self.__center

    @property
    def radius(self):
        """
        The angular radius of the region.

        Returns
        -------
        radius: astropy.units.Quantity
        """
        return self.__radius

    def contains(self, other: Any):
        """
        Check if this ConeRegion contains another ConeRegion. This region must
        be fully outside the other for this function to return True. Functionally,
        this means a region does not contain itself.

        Parameters
        ----------
        other: :py:class:`opencosmo.spatial.ConeRegion`
        """
        return self._contains(other)

    def intersects(self, other: Any):
        """
        Check if this ConeRegion intersects with another cone region. Regions which
        touch at a single point but do not cross will return False.

        Parameters
        ----------
        other: :py:class:`opencosmo.spatial.ConeRegion`
            The other region to query.

        Return
        ------
        intersects: bool
            Whether the two regions intersect
        """
        return self._intersects(other)

    def get_healpix_intersections(self, nside: int):
        phi = self.center.ra.to(u.radian).value
        dec = self.center.dec.to(u.radian).value
        # SkyCoords casts negative RAs to their positive equivalent
        # declination of north pole is +pi / 2
        # Theta of north pole is 0
        theta = np.pi / 2 - dec

        vec = ang2vec(theta, phi)
        radius = self.radius.to(u.rad).value
        return query_disc(nside, vec, radius, inclusive=True, nest=True)

    @singledispatchmethod
    def _intersects(self, other: Any):
        raise ValueError(f"Expected a 2D Sky Region but recieved {type(other)}")

    @singledispatchmethod
    def _contains(self, other: Any):
        raise ValueError(f"Expected a 2D Sky Region but recieved {type(other)}")

    @_contains.register
    def _(self, coords: SkyCoord):  # coords are assumed to be in radians
        seps = self.__center.separation(coords)
        return seps < self.__radius


@ConeRegion._contains.register  # type: ignore
def _(self, other: ConeRegion):
    dtheta = self.center.separation(other.center)
    return self.radius > (dtheta + other.radius)


@ConeRegion._intersects.register  # type: ignore
def _(self, other: ConeRegion):
    dtheta = self.center.separation(other.center)
    return dtheta < (self.radius + other.radius)


class HealPixRegion:
    def __init__(self, idxs: NDArray[np.int_], nside: int):
        self.__idxs = idxs
        self.__nside = nside

    def into_base_convention(self, *args, **kwargs):
        return self

    def into_model(self):
        return HealPixRegionModel(pixels=self.__idxs, nside=self.nside)

    @property
    def pixels(self):
        return self.__idxs

    @property
    def nside(self):
        return self.__nside

    def contains(self, other: Any):
        return False

    def intersects(self, other: Any):
        return self._intersects(other)

    @singledispatchmethod
    def _intersects(self, other: Any):
        raise ValueError(f"Expected a 2D Sky Region but recieved {type(other)}")

    @_intersects.register
    def _(self, other: ConeRegion):
        intersections = other.get_healpix_intersections(self.__nside)
        return np.any(np.isin(self.__idxs, intersections))


@HealPixRegion._intersects.register  # type: ignore
def _(self, other: HealPixRegion):
    self_pixels = self.pixels
    other_pixels = other.pixels
    return np.any(np.isin(self_pixels, other_pixels))


class FullSkyRegion:
    def __init__(self):
        pass

    def intersects(self, other: Any):
        return True

    @singledispatchmethod
    def _intersects(self, other: Any):
        raise ValueError(f"Expected a 2D Sky Region but recieved {type(other)}")

    @_intersects.register
    def _(self, other: ConeRegion):
        return True

    @_intersects.register
    def _(self, coords: SkyCoord):
        return True

    def contains(self, other: Any):
        return self._contains(other)

    @singledispatchmethod
    def _contains(self, other: Any):
        raise ValueError(f"Expected a 2D Sky Region but recieved {type(other)}")

    @_contains.register
    def _(self, other: ConeRegion):
        return True

    @_contains.register
    def _(self, coords: SkyCoord):
        return True

    def into_model(self):
        return None


@FullSkyRegion._contains.register  # type: ignore
def _(self, other: FullSkyRegion):
    return False


@FullSkyRegion._intersects.register  # type: ignore
def _(self, other: FullSkyRegion):
    return False


class BoxRegion:
    """
    A region representing a 3-dimensional box of arbitrary length, width, and depth. A
    BoxRegion can be used to query snapshot data. BoxRegions can be constructed with
    :py:meth:`opencosmo.make_box`

    When used in a spatial query, the values defining a box region will always be
    interpreted in the same unit convention as the dataset it is being used to query.
    This means two queries on the same dataset with the same box region can return
    different results if the unit convention changes between them.
    """

    def __init__(self, center: Point3d, halfwidths: BoxSize):
        self.__center = center
        self.__halfwidths = halfwidths
        self.__bounds: list[tuple[float, float]] = [
            (c - hw, c + hw) for c, hw in zip(self.__center, self.__halfwidths)
        ]

    def __repr__(self):
        return f"Box with bounds {self.bounds}"

    def into_model(self) -> BoxRegionModel:
        p1 = tuple(b[0] for b in self.bounds)
        p2 = tuple(b[1] for b in self.bounds)
        return BoxRegionModel(p1=p1, p2=p2)

    def bounding_box(self) -> BoxRegion:
        return self

    def into_base_convention(
        self,
        unit_handler: UnitHandler,
        columns: Iterable[str],
        from_: UnitConvention,
        unit_kwargs: dict[str, Any] = {},
    ):
        center = {col: dim for col, dim in zip(columns, self.__center)}
        halfwidth = {col: dim for col, dim in zip(columns, self.__halfwidths)}

        new_center = tuple(
            v.value
            for v in unit_handler.into_base_convention(center, unit_kwargs).values()
        )
        new_halfwidth = tuple(
            v.value
            for v in unit_handler.into_base_convention(halfwidth, unit_kwargs).values()
        )

        return BoxRegion(new_center, new_halfwidth)

    @property
    def bounds(self):
        """
        The bounds of this region in the form [(min,max) ...]

        Returns
        -------
        bounds: list[tuple(float,float), ....]
        """
        return self.__bounds

    def contains(self, other: Any) -> bool | np.ndarray:
        """
        Check if this box region contains another region or set of points.
        Points should be passed in as a numpy array with shape (3, n_points). Note
        that this method requires that the bounds of the test region be fully inside the
        other region. Functionally this means that a region does not contain itself.

        Parameters
        ----------
        other: BoxRegion | np.ndarray
            The region or points to check

        Returns
        -------
        contained: bool or np.ndarray[bool]
            Whether the region or points are contained in the region

        Raises
        ------
        ValueError
            If the input not a region or points
        """
        return self._contains(other)

    @singledispatchmethod
    def _contains(self, other):
        raise ValueError(f"Expected a 3D region, got {type(other)}")

    @_contains.register
    def _(self, coords: np.ndarray):
        if coords.shape[0] != 3:
            raise ValueError("Expected a coordinate array!")

        mask = np.ones(coords.shape[1], dtype=bool)
        for bound, col in zip(self.bounds, coords):
            mask &= (col > bound[0]) & (col < bound[1])

        return mask

    def intersects(self, other: Region) -> bool:
        """
        Check if this BoxRegion intersects another BoxRegion. This function
        returns true of the regions intersect in any way, including if one
        contains the other. Regions that share a bound but do not cross
        will return False.

        Parameters
        ----------
        other: BoxRegion
            The region to check

        Returns
        -------
        intersects: bool
            Whether the regions intersect

        Raises
        ------
        ValueError:
            If the input is not a BoxRegion

        """
        if not isinstance(other, BoxRegion):
            raise ValueError(f"Expected a 3D region, but got {type(other)}")

        for b1, b2 in zip(self.bounds, other.bounds):
            if b1[0] > b2[1] or b1[1] < b2[0]:
                return False
        return True


@BoxRegion._contains.register  # type: ignore
def _(self, other: BoxRegion):
    for b1, b2 in zip(self.bounds, other.bounds):
        if b1[0] > b2[0] or b1[1] < b2[1]:
            return False
    return True



from __future__ import annotations

from typing import TYPE_CHECKING, Type

import opencosmo as oc
from opencosmo.collection.simulation import SimulationCollection
from opencosmo.collection.structure.io import validate_linked_groups
from opencosmo.io.io import FILE_TYPE

if TYPE_CHECKING:
    from pathlib import Path

    from opencosmo.collection.protocols import Collection
    from opencosmo.io.io import OpenTarget


def open_simulation_files(**paths: Path) -> SimulationCollection:
    """
    Open multiple files and return a simulation collection. The data
    type of every file must be the same.

    Parameters
    ----------
    paths : str or Path
        The paths to the files to open.

    Returns
    -------
    SimulationCollection

    """
    datasets: dict[str, oc.Dataset] = {}
    for key, path in paths.items():
        dataset = oc.open(path)
        if not isinstance(dataset, oc.Dataset):
            raise ValueError("All datasets must be of the same type.")
    dtypes = set(dataset for dataset in datasets.values())
    if len(dtypes) != 1:
        raise ValueError("All datasets must be of the same type.")
    return SimulationCollection(datasets)


def get_collection_type(
    targets: list[OpenTarget], file_types: list[FILE_TYPE]
) -> Type["Collection"]:
    """
    If there are multiple files, determine their collection type. There are
    three options we support at present:

    1. All files contain a single lightcone dataset, all of the same type
    2. The files contain a single non-lightcone datatype.
    3. The files are linked together into a structure collection
    """

    handles_by_type = {target.data_type: target.group for target in targets}
    is_lightcone = [target.header.file.is_lightcone for target in targets]
    unique_data_types = set(handles_by_type.keys())
    unique_file_types = set(file_types)
    if len(unique_data_types) == 1 and all(is_lightcone):
        return oc.Lightcone

    elif unique_file_types == {FILE_TYPE.STRUCTURE_COLLECTION}:
        validate_linked_groups(handles_by_type)
        return oc.StructureCollection

    elif len(unique_data_types) == 1 and all(not il for il in is_lightcone):
        return oc.SimulationCollection

    elif FILE_TYPE.HALO_PROPERTIES in unique_file_types and set(
        [FILE_TYPE.HALO_PARTICLES, FILE_TYPE.SOD_BINS, FILE_TYPE.GALAXY_PROPERTIES]
    ).intersection(unique_file_types):
        validate_linked_groups(handles_by_type)
        return oc.StructureCollection

    elif unique_file_types == {FILE_TYPE.GALAXY_PROPERTIES, FILE_TYPE.GALAXY_PARTICLES}:
        return oc.StructureCollection

    elif (
        len(unique_file_types) == 1
        and unique_file_types.pop() == FILE_TYPE.STRUCTURE_COLLECTION
    ):
        validate_linked_groups(handles_by_type)
        return oc.StructureCollection
    else:
        raise ValueError("Invalid combination of files")



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import h5py


def combine_upwards(
    counts: np.ndarray, factor: int, level: int, target: h5py.File
) -> h5py.File:
    """
    Given a count of the number of items in each region at a given level, write the
    starts and sizes to and hdf5 dataset and recursively work upwards until the
    top level of the index is reached. The index should verify that the length of
    the initial array it recieves is the correct length for the given level.
    """
    if (len(counts) % factor**level) != 0:
        raise ValueError("Recieved invalid number of counts!")

    group = target.require_group(f"level_{level}")
    new_starts = np.insert(np.cumsum(counts, dtype=np.int32), 0, 0)[:-1]
    counts = counts.astype(np.int32)  # This should be fixed
    group.create_dataset("start", data=new_starts)
    group.create_dataset("size", data=counts)

    if level > 0:
        new_counts = counts.reshape(-1, factor).sum(axis=1)
        return combine_upwards(new_counts, factor, level - 1, target)

    return target



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Iterable, Optional

import numpy as np
from deprecated import deprecated

from opencosmo import io
from opencosmo.collection import lightcone as lc
from opencosmo.collection.structure import structure as sc
from opencosmo.dataset.handler import Hdf5Handler
from opencosmo.index import SimpleIndex

if TYPE_CHECKING:
    from pathlib import Path

    import h5py

    from opencosmo import dataset as d
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.index import DataIndex

ALLOWED_LINKS = {  # h5py.Files that can serve as a link holder and
    "halo_properties": ["halo_particles", "halo_profiles", "galaxy_properties"],
    "galaxy_properties": ["galaxy_particles"],
}


def remove_empty(dataset):
    metadata = dataset.get_metadata()
    mask = np.ones(len(dataset), dtype=bool)
    for name, col in metadata.items():
        if "size" in name:
            mask &= col != 0
        elif "idx" in name:
            mask &= col != -1

    if not mask.all():
        dataset = dataset.take_rows(np.where(mask)[0])
    return dataset


@deprecated(
    version="0.8",
    reason="oc.open_linked_files is deprecated and will be removed in version 1.0. "
    "Please use oc.open instead",
)
def open_linked_files(*files: Path, **load_kwargs: bool):
    """
    **WARNING: THIS METHOD IS DEPCREATED AND WILL BE REMOVED IN A FUTURE
    VERSION. PLEASE USE** :py:meth:`opencosmo.open`

    Open a collection of files that are linked together, such as a
    properties file and a particle file.

    """
    if len(files) == 1 and isinstance(files[0], list):
        return open_linked_files(*files[0])

    return io.io.open(*files, **load_kwargs)


def validate_linked_groups(groups: dict[str, h5py.Group]):
    if "halo_properties" in groups:
        if "data_linked" not in groups["halo_properties"].keys():
            raise ValueError(
                "File appears to be a structure collection, but does not have links!"
            )
    elif "galaxy_properties" in groups:
        if "data_linked" not in groups["galaxy_properties"].keys():
            raise ValueError(
                "File appears to be a structure collection, but does not have links!"
            )
    if len(groups) == 1:
        raise ValueError("Structure collections must have more than one dataset")


def get_linked_datasets(
    linked_files_by_type: dict[str, h5py.File | h5py.Group],
    header: OpenCosmoHeader,
):
    targets = {}
    for dtype, pointer in linked_files_by_type.items():
        if "data" not in pointer.keys():
            targets.update(
                {
                    k: io.io.OpenTarget(pointer[k], header)
                    for k in pointer.keys()
                    if k != "header"
                }
            )
        else:
            targets.update({dtype: io.io.OpenTarget(pointer, header)})
    datasets = {
        dtype: io.io.open_single_dataset(target, bypass_lightcone=True, bypass_mpi=True)
        for dtype, target in targets.items()
    }
    return datasets


def build_structure_collection(targets: list[io.io.OpenTarget], ignore_empty: bool):
    link_sources = defaultdict(list)
    link_targets: dict[str, dict[str, list[d.Dataset | sc.StructureCollection]]] = (
        defaultdict(lambda: defaultdict(list))
    )
    for target in targets:
        if target.data_type == "halo_properties":
            link_sources["halo_properties"].append(target)
        elif target.data_type == "galaxy_properties":
            link_sources["galaxy_properties"].append(target)
        elif target.data_type.startswith("halo"):
            dataset = io.io.open_single_dataset(
                target, bypass_lightcone=True, bypass_mpi=True
            )
            name = target.group.name.split("/")[-1]
            if not name:
                name = target.data_type
            elif name.startswith("halo_properties"):
                name = name[16:]
            link_targets["halo_targets"][name].append(dataset)
        elif target.data_type.startswith("galaxy"):
            dataset = io.io.open_single_dataset(
                target, bypass_lightcone=True, bypass_mpi=True
            )
            name = target.group.name.split("/")[-1]
            if not name:
                name = target.data_type
            elif name.startswith("galaxy_properties"):
                name = name[18:]
            link_targets["galaxy_targets"][name].append(dataset)
        else:
            raise ValueError(
                f"Unknown data type for structure collection {target.data_type}"
            )

    if (
        len(link_sources["halo_properties"]) > 1
        or len(link_sources["galaxy_properties"]) > 1
    ):
        raise NotImplementedError(
            "Opening structure collections that span multiple redshifts is not currently supported"
        )
        # Potentially a lightcone structure collection
        collections = {}
        sources_by_step, targets_by_step = __sort_by_step(link_sources, link_targets)
        if set(sources_by_step.keys()) != set(targets_by_step.keys()):
            raise ValueError("Datasets are not the same across all lightcone steps!")
        for step, sources in sources_by_step.items():
            halo_properties = sources.get("halo_properties")
            galaxy_properties = sources.get("galaxy_properties")
            targets = targets_by_step[step]
            collection = __build_structure_collection(
                halo_properties, galaxy_properties, targets, ignore_empty
            )
            collections[step] = collection

        expected_datasets = set(next(iter(collections.values())).keys())
        for collection in collections.values():
            if set(collection.keys()) != expected_datasets:
                raise ValueError(
                    "All structure collections in a lightcone must have the same set of datasets"
                )
        return lc.Lightcone(collections)

    halo_properties_target = None
    galaxy_properties_target = None
    if link_sources["halo_properties"]:
        halo_properties_target = link_sources["halo_properties"][0]
    if link_sources["galaxy_properties"]:
        galaxy_properties_target = link_sources["galaxy_properties"][0]

    input_link_targets: dict[str, dict[str, d.Dataset | sc.StructureCollection]] = (
        defaultdict(dict)
    )
    for source_type, source_targets in link_targets.items():
        if any(len(ts) > 1 for ts in source_targets.values()):
            raise ValueError("Found more than one linked file of a given type!")
        input_link_targets[source_type] = {
            key: t[0] for key, t in source_targets.items()
        }

    return __build_structure_collection(
        halo_properties_target,
        galaxy_properties_target,
        input_link_targets,
        ignore_empty,
    )


def __sort_by_step(link_sources: dict[str, list[io.io.OpenTarget]], link_targets):
    sources_by_step: dict[int, dict[str, io.io.OpenTarget]] = defaultdict(dict)
    targets_by_step: dict[int, dict[str, dict[str, d.Dataset]]] = defaultdict(
        lambda: defaultdict(dict)
    )
    for source_name, sources in link_sources.items():
        for source in sources:
            if not source.header.file.is_lightcone:
                raise ValueError(
                    "Recived multiple source datasets of a single type, but not all are lightcone datasets!"
                )
            sources_by_step[source.header.file.step][source_name] = source
    for target_type, targets_ in link_targets.items():
        for target_name, targets in targets_.items():
            for target in targets:
                if not target.header.file.is_lightcone:
                    raise ValueError(
                        "Recived multiple datasets of a single type, but not all are lightcone datasets!"
                    )
                targets_by_step[target.header.file.step][target_type][target_name] = (
                    target
                )

    return sources_by_step, targets_by_step


def __build_structure_collection(
    halo_properties_target: Optional[io.io.OpenTarget],
    galaxy_properties_target: Optional[io.io.OpenTarget],
    link_targets: dict[str, dict[str, d.Dataset | sc.StructureCollection]],
    ignore_empty: bool,
):
    if galaxy_properties_target is not None and "galaxy_targets" in link_targets:
        source_dataset = io.io.open_single_dataset(
            galaxy_properties_target,
            metadata_group="data_linked",
            bypass_lightcone=True,
            bypass_mpi=halo_properties_target is not None,
        )
        if ignore_empty and halo_properties_target is None:
            source_dataset = remove_empty(source_dataset)
        collection = sc.StructureCollection(
            source_dataset,
            source_dataset.header,
            link_targets["galaxy_targets"],
        )
        if halo_properties_target is not None:
            link_targets["halo_targets"]["galaxy_properties"] = collection
        else:
            return collection

    if (
        halo_properties_target is not None
        and galaxy_properties_target is not None
        and "galaxy_targets" not in link_targets
    ):
        galaxy_properties = io.io.open_single_dataset(
            galaxy_properties_target, bypass_lightcone=True
        )
        link_targets["halo_targets"]["galaxy_properties"] = galaxy_properties

    if halo_properties_target is not None and link_targets["halo_targets"]:
        source_dataset = io.io.open_single_dataset(
            halo_properties_target, metadata_group="data_linked", bypass_lightcone=True
        )
        if ignore_empty:
            source_dataset = remove_empty(source_dataset)

        return sc.StructureCollection(
            source_dataset,
            source_dataset.header,
            link_targets["halo_targets"],
        )



# ruff: noqa: TC001 TC003
from datetime import date
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Optional

import astropy.cosmology.units as cu  # type: ignore
import astropy.units as u  # type: ignore
import numpy as np
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)

from opencosmo.units import UnitConvention

from .cosmology import CosmologyParameters
from .diffsky import DiffskyVersionInfo
from .units import register_units


def empty_string_to_none(v):
    if isinstance(v, str) and v == "":
        return None
    return v


class HaccSimulationParameters(BaseModel):
    ACCESS_PATH: ClassVar[str] = "simulation"
    model_config = ConfigDict(frozen=True)

    box_size: float = Field(ge=0, description="Size of the simulation box (Mpc/h)")
    z_ini: float = Field(ge=0.01, description="Initial redshift of the simulation")
    z_end: float = Field(ge=0.0, description="Final redshift of the simulation")
    n_gravity: Optional[int] = Field(
        ge=2,
        description="Number of gravity-only particles (per dimension). "
        "In hydrodynamic simulations, this parameter will not be set.",
    )
    n_steps: int = Field(ge=1, description="Number of time steps")
    pm_grid: int = Field(ge=2, description="Number of grid points (per dimension)")
    offset_gravity_ini: Optional[float] = Field(
        description="Lagrangian offset for gravity-only particles"
    )

    @model_validator(mode="before")
    @classmethod
    def empty_string_to_none(cls, data):
        if isinstance(data, dict):
            return {k: empty_string_to_none(v) for k, v in data.items()}
        return data

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def step_zs(self) -> list[float]:
        """
        Get the redshift of the steps in this simulation. Outputs such that
        redshift[step_number] returns the redshift for that step. Keep in
        mind that steps go from high z -> low z.
        """
        a_ini = 1 / (1 + self.z_ini)
        a_end = 1 / (1 + self.z_end)
        # Steps are evenly spaced in a with step zero corresponding to the first step
        # after the initial conditions
        step_as = np.linspace(a_ini, a_end, self.n_steps + 1)[1:]
        return np.round(1 / step_as - 1, 3).tolist()  # type: ignore


HaccGravityOnlySimulationParameters = HaccSimulationParameters


def subgrid_alias_generator(name: str) -> str:
    return f"subgrid_{name}"


class HaccHydroSimulationParameters(HaccSimulationParameters):
    model_config = ConfigDict(frozen=True)
    n_gas: int = Field(
        description="Number of gas particles (per dimension)", alias="n_bar"
    )
    n_dm: int = Field(
        ge=2, description="Number of dark matter particles (per dimension)"
    )
    offset_gas_ini: float = Field(
        description="Lagrangian offset for gas particles", alias="offset_bar_ini"
    )
    offset_dm_ini: float = Field(
        description="Lagrangian offset for dark matter particles"
    )
    agn_kinetic_eps: float = Field(
        description="AGN feedback efficiency", alias="subgrid_agn_kinetic_eps"
    )
    agn_kinetic_jet_vel: float = Field(
        description="AGN feedback velocity", alias="subgrid_agn_kinetic_jet_vel"
    )
    agn_nperh: float = Field(
        description="AGN sphere of influence", alias="subgrid_agn_nperh"
    )
    agn_seed_mass: float = Field(
        description="AGN seed mass (Msun / h)", alias="subgrid_agn_seed_mass"
    )
    wind_egy_w: float = Field(
        description="Wind mass loading factor", alias="subgrid_wind_egy_w"
    )
    wind_kappa_w: float = Field(
        description="Wind velocity", alias="subgrid_wind_kappa_w"
    )


class CosmoToolsParameters(BaseModel):
    model_config = ConfigDict(frozen=True)
    ACCESS_PATH: ClassVar[str] = "cosmotools"
    cosmotools_steps: frozenset[int]
    fof_linking_length: float
    fof_pmin: int
    sod_pmin: int
    sod_delta_crit: float
    sod_concentration_pmin: int
    sodbighaloparticles_pmin: int
    profiles_nbins: int
    galaxy_dbscan_neighbors: Optional[int]
    galaxy_aperture_radius: Optional[int]
    galaxy_pmin: Optional[int]

    @model_validator(mode="before")
    @classmethod
    def empty_string_to_none(cls, data):
        if isinstance(data, dict):
            data = {k: empty_string_to_none(v) for k, v in data.items()}
        return data

    @field_serializer("cosmotools_steps")
    def serialize_steps(self, steps) -> list[int]:
        return list(steps)


class ReformatParameters(BaseModel):
    model_config = ConfigDict(frozen=True)
    ACCESS_PATH: ClassVar[str] = "reformat"
    cosmotools_lc_path: Optional[Path] = None
    cosmotools_path: Path
    indat_path: Path
    is_hydro: bool
    lightcone_analysis_path_pattern: Optional[str] = None
    machine: str
    mass_threshold_sodbighaloparticles: Optional[float] = None
    mass_threshold_sodpropertybins: Optional[float] = None
    max_level: int = 0
    max_level_lc: Optional[frozenset[tuple[int, int]]] = None
    npart_threshold_galaxyproperties: Optional[int] = None
    output_lc_path_pattern: Optional[str] = None
    rearrange_output_path_pattern: str
    rearrange_output_lc_path_pattern: Optional[str] = None
    simulation_date: date
    simulation_name: str
    snapshot_analysis_path_pattern: Optional[str] = None
    temporary_path: Optional[Path] = None

    @field_serializer(
        "cosmotools_lc_path",
        "cosmotools_path",
        "indat_path",
        "temporary_path",
        "lightcone_analysis_path_pattern",
        "output_lc_path_pattern",
        "rearrange_output_lc_path_pattern",
        "snapshot_analysis_path_pattern",
    )
    def handle_path(self, v):
        if v is None:
            return ""
        return str(v)

    @field_serializer("max_level_lc")
    def serialize_max_level(self, value):
        if value is None:
            return
        return list(value)

    @field_serializer("simulation_date")
    def handle_date(self, v):
        return v.isoformat()

    @field_validator("is_hydro", mode="before")
    @classmethod
    def numpy_bool_to_base(cls, value):
        if isinstance(value, np.bool_):
            return bool(value)
        return value

    @model_validator(mode="before")
    @classmethod
    def empty_string_to_none(cls, data):
        if isinstance(data, dict):
            data = {k: empty_string_to_none(v) for k, v in data.items()}
        return data


class LightconeParams(BaseModel):
    model_config = ConfigDict(frozen=True)
    ACCESS_PATH: ClassVar[str] = "lightcone"
    z_range: Optional[tuple[float, float]] = None

    @model_validator(mode="before")
    @classmethod
    def empty_string_to_none(cls, data):
        if isinstance(data, dict):
            return {k: empty_string_to_none(v) for k, v in data.items()}
        return data


ORIGIN_PARAMETERS = {
    "required": {
        "simulation/parameters": HaccHydroSimulationParameters
        | HaccGravityOnlySimulationParameters,
        "simulation/cosmotools": CosmoToolsParameters,
        "simulation/cosmology": CosmologyParameters,
    },
    "optional": {"reformat_hacc/config": ReformatParameters},
}

DATATYPE_PARAMETERS: dict[str, dict[str, type[BaseModel]]] = {
    "halo_properties": {},
    "galaxy_properties": {},
    "halo_particles": {},
    "galaxy_particles": {},
    "halo_profiles": {},
    "diffsky_fits": {"diffsky_versions": DiffskyVersionInfo},
}

register_units(
    HaccSimulationParameters, "box_size", u.Mpc / cu.littleh, UnitConvention.SCALEFREE
)
register_units(
    HaccHydroSimulationParameters,
    "box_size",
    u.Mpc / cu.littleh,
    UnitConvention.SCALEFREE,
)
register_units(
    HaccHydroSimulationParameters,
    "agn_seed_mass",
    u.Msun / cu.littleh,
    UnitConvention.SCALEFREE,
)



from __future__ import annotations

from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, Iterable, Optional
from weakref import ref

import numpy as np

from opencosmo.column.cache import ColumnCache
from opencosmo.index import ChunkedIndex, SimpleIndex
from opencosmo.index.take import take
from opencosmo.io.schemas import DatasetSchema
from opencosmo.mpi import get_comm_world

if TYPE_CHECKING:
    import h5py

    from opencosmo.dataset.column import DerivedColumn
    from opencosmo.dataset.state import DatasetState
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.index import DataIndex


class Hdf5Handler:
    """
    Handler for opencosmo.Dataset
    """

    def __init__(
        self,
        group: h5py.Group,
        index: DataIndex,
        cache: ColumnCache,
        metadata_group: Optional[h5py.Group],
    ):
        self.__index = index
        self.__group = group
        self.__cache = cache
        self.__metadata_group = metadata_group

    @classmethod
    def from_group(
        cls,
        group: h5py.Group,
        index: Optional[DataIndex] = None,
        metadata_group: Optional[h5py.Group] = None,
    ):
        if not group.name.endswith("data"):
            raise ValueError("Expected a data group")
        lengths = set(len(ds) for ds in group.values())
        if len(lengths) > 1:
            raise ValueError("Not all columns are the same length!")

        if index is None:
            index = ChunkedIndex.from_size(lengths.pop())

        colnames = group.keys()
        if metadata_group is not None:
            colnames = chain(colnames, metadata_group.keys())

        return Hdf5Handler(group, index, ColumnCache.empty(), metadata_group)

    def register(self, state: DatasetState, derived: dict[str, DerivedColumn]):
        columns = set(state.columns).difference(derived)
        for dc in derived.values():
            columns |= dc.requires()

        self.__cache.register_column_group(
            id(state), columns.union(self.metadata_columns)
        )

    def deregister(self, state_id: int):
        self.__cache.deregister_column_group(state_id)

    def take(self, other: DataIndex, sorted: Optional[np.ndarray] = None):
        if len(other) == 0:
            return Hdf5Handler(
                self.__group, other, ColumnCache.empty(), self.__metadata_group
            )

        if sorted is not None:
            return self.__take_sorted(other, sorted)

        new_index = take(self.__index, other)
        new_cache = self.__cache.take(other)
        return Hdf5Handler(self.__group, new_index, new_cache, self.__metadata_group)

    def __take_sorted(self, other: DataIndex, sorted: np.ndarray):
        if len(sorted) != len(self.__index):
            raise ValueError("Sorted index has the wrong length!")
        new_indices = other.get_data(sorted)

        new_raw_index = self.__index.into_array()[new_indices]
        new_index = SimpleIndex(np.sort(new_raw_index))

        new_cache_index = SimpleIndex(new_indices)
        new_cache = self.__cache.take(new_cache_index)

        return Hdf5Handler(self.__group, new_index, new_cache, self.__metadata_group)

    @property
    def data(self):
        return self.__group

    @property
    def index(self):
        return self.__index

    @property
    def columns(self):
        return self.__group.keys()

    @property
    def metadata_columns(self):
        if self.__metadata_group is None:
            return []
        return self.__metadata_group.keys()

    @cached_property
    def descriptions(self):
        return {
            colname: column.attrs.get("description")
            for colname, column in self.__group.items()
        }

    def mask(self, mask):
        idx = SimpleIndex(np.where(mask)[0])
        return self.take(idx)

    def __len__(self) -> int:
        first_column_name = next(iter(self.__group.keys()))
        return self.__group[first_column_name].shape[0]

    def __enter__(self):
        return self

    def __exit__(self, *exec_details):
        self.__group = None
        return self.__file.close()

    def make_schema(
        self,
        columns: Iterable[str],
        header: Optional[OpenCosmoHeader] = None,
    ) -> DatasetSchema:
        groups = {}
        columns = set(columns)
        data_columns = [f"data/{n}" for n in columns]
        groups["data"] = self.__group

        cached_data = {
            f"data/{n}": data for n, data in self.__cache.get_columns(columns).items()
        }

        if self.metadata_columns:
            assert self.__metadata_group is not None
            group_name = self.__metadata_group.name.split("/")[-1]
            metadata_columns = [f"{group_name}/{n}" for n in self.metadata_columns]
            cached_data |= {
                f"{group_name}/{n}": data
                for n, data in self.__cache.get_columns(self.metadata_columns).items()
            }
            groups[group_name] = self.__metadata_group
        else:
            metadata_columns = []
        return DatasetSchema.make_schema(
            groups,
            data_columns + metadata_columns,
            self.__index,
            cached_data,
            header,
        )

    def get_data(self, columns: Iterable[str]) -> dict[str, np.ndarray]:
        """ """
        if self.__group is None:
            raise ValueError("This file has already been closed")
        cached_data = self.__cache.get_columns(columns)
        remaining = set(columns).difference(cached_data.keys())
        new_data = {}

        for colname in remaining:
            new_data[colname] = self.__index.get_data(self.__group[colname])
        if new_data:
            self.__cache = self.__cache.with_data(new_data)

        data = new_data | cached_data

        # Ensure order is preserved
        return {name: data[name] for name in columns}

    def get_metadata(self, columns: Iterable[str]) -> Optional[dict[str, np.ndarray]]:
        if self.__metadata_group is None:
            return None
        if not columns:
            columns = self.metadata_columns

        data = {}
        for colname in columns:
            data[colname] = self.__index.get_data(self.__metadata_group[colname])

        return data

    def take_range(self, start: int, end: int, indices: np.ndarray) -> np.ndarray:
        if start < 0 or end > len(indices):
            raise ValueError("Indices out of range")
        return indices[start:end]



from __future__ import annotations

import shutil
import subprocess
import sys
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .specs import DependencySpec


def install_conda_forge(packages: dict[str, Optional[str]]):
    conda_executable = shutil.which("conda")

    prefix = sys.prefix
    if conda_executable is None:
        raise FileNotFoundError("Unable to locate the conda executable")

    command = f"{conda_executable} install --prefix {prefix} -y -c conda-forge"
    for name, version in packages.items():
        command += f" {name}"
        if version is not None:
            command += f"={version}"

    run_install_command(command)


def install_pip(packages: dict[str, Optional[str]], dev: bool = False):
    if dev:
        command = "uv pip install"
    else:
        command = f"{sys.executable} -m pip install"
    for name, version in packages.items():
        command += f" {name}"
        if version is not None:
            command += f"=={version}"
    run_install_command(command)


def install_github(name: str, version: str, dep_spec: dict[str, DependencySpec]):
    repo_url = dep_spec[name].repo
    assert repo_url is not None
    command = f"{sys.executable} -m pip install git+{repo_url}"
    commit = version.split("+g")[-1]
    command += f"@{commit}"

    run_install_command(command)


def run_install_command(command: str):
    print(command)
    _ = subprocess.run(command.split(" "))



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Optional

import numpy as np

from opencosmo.index import SimpleIndex

if TYPE_CHECKING:
    import astropy.units as u
    from numpy.typing import NDArray

    from opencosmo.index import DataIndex


class InMemoryColumnHandler:
    def __init__(
        self,
        columns: dict[str, NDArray | u.Quantity],
        descriptions: dict[str, Optional[str]] = {},
    ):
        self.__columns = columns
        self.__descriptions = descriptions
        lengths = set(len(c) for c in columns.values())
        if len(lengths) > 1:
            raise ValueError(
                "Tried to instantiate an InMemoryColumnHandler with columns of different lengths!"
            )
        elif len(lengths) == 0:
            lengths = set([0])
        self.__length = lengths.pop()

    @property
    def descriptions(self):
        return self.__descriptions

    @classmethod
    def empty(cls):
        return InMemoryColumnHandler({}, {})

    def __len__(self):
        return self.__length

    def with_columns(self, columns: Iterable[str]):
        new_columns = {
            key: self.__columns[key] for key in columns if key in self.__columns
        }
        return InMemoryColumnHandler(new_columns, self.__descriptions)

    def get_rows(self, index: np.ndarray):
        if self.__length == 0:
            return InMemoryColumnHandler({}, {})

        if index.dtype == np.bool_ and len(index) > self.__length:
            raise ValueError("Received a boolean index that is longer than the data!")

        if np.max(index) > self.__length:
            raise ValueError("Tried to get a rows that are not in this column!")

        new_data = {key: data[index] for key, data in self.__columns.items()}
        return InMemoryColumnHandler(new_data, self.__descriptions)

    def take_range(self, start: int, end: int):
        if self.__length == 0:
            return self
        elif start >= self.__length or end > self.__length:
            raise ValueError(
                "The requested range is outside the range of this dataset!"
            )
        new_data = {key: data[start:end] for key, data in self.__columns.items()}
        return InMemoryColumnHandler(new_data, self.__descriptions)

    def take(self, index: DataIndex, sorted: Optional[np.ndarray] = None):
        if len(index) == 0:
            return InMemoryColumnHandler.empty()
        if len(self) == 0:
            return self
        if sorted is not None:
            return self.__take_sorted(index, sorted)
        new_data = {key: index.get_data(value) for key, value in self.__columns.items()}
        return InMemoryColumnHandler(new_data, self.__descriptions)

    def __take_sorted(self, index: DataIndex, sorted_by: np.ndarray):
        if len(sorted_by) != len(self):
            raise ValueError()
        new_indices = index.get_data(sorted_by)
        new_index = SimpleIndex(np.sort(new_indices))
        return self.take(new_index)

    def keys(self):
        return self.__columns.keys()

    def get_data(self, keys: str | Iterable[str]):
        if isinstance(keys, str):
            keys = [keys]
        missing = set(keys) - set(self.keys())
        if missing:
            raise ValueError(f"Requested unknown columns {missing}")
        return {name: self.__columns[name] for name in keys}

    def with_new_column(
        self,
        name: str,
        column: np.ndarray | u.Quantity,
        description: Optional[str] = None,
    ):
        if self.__columns and len(column) != len(self):
            raise ValueError("Tried to add an in-memory column with the wrong length!")
        new_columns = {**self.__columns, name: column}
        new_descriptions = {**self.__descriptions, name: description}
        return InMemoryColumnHandler(new_columns, new_descriptions)

    def columns(self):
        yield from self.__columns.items()



from __future__ import annotations

from functools import partial, reduce
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterable, Mapping, Optional
from warnings import warn

import numpy as np

import opencosmo as oc
from opencosmo.collection.structure import evaluate
from opencosmo.collection.structure import io as sio
from opencosmo.index import ChunkedIndex, SimpleIndex
from opencosmo.io.schemas import StructCollectionSchema

if TYPE_CHECKING:
    import astropy
    import astropy.units as u

    from opencosmo.dataset.column import DerivedColumn
    from opencosmo.index import DataIndex
    from opencosmo.io import io
    from opencosmo.parameters import HaccSimulationParameters
    from opencosmo.spatial.protocols import Region


def filter_source_by_dataset(
    dataset: oc.Dataset,
    source: oc.Dataset,
    header: oc.header.OpenCosmoHeader,
    *masks,
) -> oc.Dataset:
    masked_dataset = dataset.filter(*masks)
    linked_column: str
    if header.file.data_type == "halo_properties":
        linked_column = "fof_halo_tag"
    elif header.file.data_type == "galaxy_properties":
        linked_column = "gal_tag"

    tags = masked_dataset.select(linked_column).data
    new_source = source.filter(oc.col(linked_column).isin(tags))
    return new_source


LINK_ALIASES = {  # Left: Name in file, right: Name in collection
    "sodbighaloparticles_star_particles": "star_particles",
    "sodbighaloparticles_dm_particles": "dm_particles",
    "sodbighaloparticles_gravity_particles": "gravity_particles",
    "sodbighaloparticles_agn_particles": "agn_particles",
    "sodbighaloparticles_gas_particles": "gas_particles",
    "sod_profile": "halo_profiles",
    "galaxyproperties": "galaxy_properties",
    "galaxyparticles_star_particles": "star_particles",
}


def create_start_size(data, start_name, size_name):
    start = data.pop(start_name, None)
    size = data.pop(size_name, None)
    if start is None:
        return None
    if isinstance(start, np.ndarray):
        return ChunkedIndex(start, size)
    if size == 0:
        return None
    return ChunkedIndex.single_chunk(start, size)


def create_idx(data, idx_name):
    idx = data.pop(idx_name, None)
    if idx is None:
        return None

    if isinstance(idx, np.ndarray):
        return SimpleIndex(idx)
    elif idx == -1:
        return None
    return SimpleIndex(np.atleast_1d(idx))


def make_links(keys, rename_galaxies=False):
    starts = list(filter(lambda key: "start" in key, keys))
    sizes = list(filter(lambda key: "size" in key, keys))
    idxs = list(filter(lambda key: "idx" in key, keys))

    starts = set(map(lambda key: key[:-6], starts))
    sizes = set(map(lambda key: key[:-5], sizes))
    idxs = set(map(lambda key: key[:-4], idxs))

    assert starts == sizes
    output = {}
    columns = {}
    for name in starts:
        output[LINK_ALIASES[name]] = partial(
            create_start_size, start_name=f"{name}_start", size_name=f"{name}_size"
        )
        columns[LINK_ALIASES[name]] = [f"{name}_start", f"{name}_size"]

    for name in idxs:
        output[LINK_ALIASES[name]] = partial(create_idx, idx_name=f"{name}_idx")
        columns[LINK_ALIASES[name]] = [f"{name}_idx"]

    if rename_galaxies and "galaxy_properties" in output:
        output["galaxies"] = output.pop("galaxy_properties")
        columns["galaxies"] = columns.pop("galaxy_properties")
    return output, columns


class LinkHandler:
    def __init__(self, links: Iterable[str], rename_galaxies=False):
        self.links, self.columns = make_links(links, rename_galaxies)

    def parse(self, data: dict[str, Any]):
        output = {}
        for name, handler in self.links.items():
            result = handler(data)
            if result is not None:
                output[name] = result
        return output


class StructureCollection:
    """
    A collection of datasets that contain both high-level properties
    and lower level information (such as particles) for structures
    in the simulation. Currently these structures include halos
    and galaxies.

    Every structure collection has a halo_properties or galaxy_properties dataset
    that contains the high-level measured attribute of the structures. Certain
    operations (e.g. :py:meth:`sort_by <opencosmo.StructureCollection.sort_by>`
    operate on this dataset.
    """

    def __init__(
        self,
        source: oc.Dataset,
        header: oc.header.OpenCosmoHeader,
        datasets: Mapping[str, oc.Dataset | StructureCollection],
        hide_source: bool = False,
        **kwargs,
    ):
        """
        Initialize a linked collection with the provided datasets and links.
        """

        self.__source = source
        self.__header = header
        self.__datasets = dict(datasets)
        self.__index = self.__source.index
        self.__hide_source = hide_source
        if isinstance(self.__datasets.get("galaxy_properties"), StructureCollection):
            self.__datasets["galaxies"] = self.__datasets.pop("galaxy_properties")

        self.__handler = LinkHandler(
            self.__source.meta_columns, "galaxies" in self.__datasets
        )

    def __repr__(self):
        structure_type = self.__header.file.data_type.split("_")[0] + "s"
        keys = list(self.keys())
        if len(keys) == 2:
            dtype_str = " and ".join(keys)
        else:
            dtype_str = ", ".join(keys[:-1]) + ", and " + keys[-1]
        return f"Collection of {structure_type} with {dtype_str}"

    def __len__(self):
        return len(self.__source)

    @classmethod
    def open(
        cls, targets: list[io.OpenTarget], ignore_empty=True, **kwargs
    ) -> StructureCollection:
        return sio.build_structure_collection(targets, ignore_empty)

    @property
    def header(self):
        return self.__header

    @property
    def dtype(self):
        structure_type = self.__header.file.data_type.split("_")[0]
        return structure_type

    @property
    def cosmology(self) -> astropy.cosmology.Cosmology:
        """
        The cosmology of the structure collection
        """
        return self.__source.cosmology

    @property
    def properties(self) -> list[str]:
        """
        The high-level properties that are available as part of the
        halo_properties or galaxy_properties dataset.
        """
        return self.__source.columns

    @property
    def redshift(self) -> float | tuple[float, float]:
        """
        For snapshots, return the redshift or redshift range
        this dataset was drawn from.

        Returns
        -------
        redshift: float | tuple[float, float]

        """
        return self.__header.file.redshift

    @property
    def simulation(self) -> HaccSimulationParameters:
        """
        Get the parameters of the simulation this dataset is drawn
        from.

        Returns
        -------
        parameters: opencosmo.parameters.HaccSimulationParameters
        """
        return self.__header.simulation

    def keys(self) -> list[str]:
        """
        Return the names of the datasets in this collection.
        """
        keys = list(self.__datasets.keys())
        if not self.__hide_source:
            keys.append(self.__source.dtype)
        return keys

    def values(self) -> list[oc.Dataset | StructureCollection]:
        """
        Return the datasets in this collection.
        """
        return [self[name] for name in self.keys()]

    def items(self) -> Generator[tuple[str, oc.Dataset | StructureCollection]]:
        """
        Return the names and datasets as key-value pairs.
        """

        for k, v in zip(self.keys(), self.values()):
            yield k, v

    def __getitem__(self, key: str) -> oc.Dataset | oc.StructureCollection:
        """
        Return the linked dataset with the given key.
        """
        if key not in self.keys():
            raise KeyError(f"Dataset {key} not found in collection.")
        elif key == self.__header.file.data_type:
            return self.__source
        ds = self.__datasets[key]
        if isinstance(ds, StructureCollection):
            idx = ds.__source.index
            if not idx.range()[0] == 0 or len(ds) != idx.range()[1]:
                return ds

        meta = self.__source.get_metadata(self.__handler.columns[key])
        indices = self.__handler.parse(meta)
        ds = self.__datasets[key]
        index = indices[key]

        return self.__datasets[key].take_rows(index)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        for dataset in self.values():
            try:
                dataset.__exit__(*args)
            except AttributeError:
                continue

    @property
    def region(self):
        return self.__source.region

    def bound(
        self, region: Region, select_by: Optional[str] = None
    ) -> StructureCollection:
        """
        Restrict this collection to only contain structures in the specified region.
        Querying will be done based on the halo  or galaxy centers, meaning some
        particles may fall outside the given region.

        See :doc:`spatial_ref` for details of how to construct regions.

        Parameters
        ----------
        region: opencosmo.spatial.Region

        Returns
        -------
        dataset: opencosmo.Dataset
            The portion of the dataset inside the selected region

        Raises
        ------
        ValueError
            If the query region does not overlap with the region this dataset resides
            in
        AttributeError:
            If the dataset does not contain a spatial index
        """

        bounded = self.__source.bound(region, select_by)
        return StructureCollection(
            bounded, self.__header, self.__datasets, self.__hide_source
        )

    def evaluate(
        self,
        func: Callable,
        dataset: Optional[str] = None,
        format: str = "astropy",
        vectorize: bool = False,
        insert: bool = True,
        **evaluate_kwargs: Any,
    ):
        """
        Iterate over the structures in this collection and apply func to each,
        collecting the results into a new column. These values will be computed
        immediately rather than lazily. If your new column can be created from a
        simple algebraic combination of existing columns, use
        :py:meth:`with_new_columns <opencosmo.StructureCollection.with_new_columns>`.

        You can substantially improve the performance of this method by specifying
        which data is actually needed to do the computation. This method will
        automatically select the requested data, avoiding reading unneeded data
        from disk. The semantics for specifying the columns is identical to
        :py:meth:`select <opencosmo.StructureCollection.select>`.

        The function passed to this method must take arguments that match the names
        of datasets that are stored in this collection. You can specify specific
        columns that are needed with keyword arguments to this function. For example:

        .. code-block:: python

            import opencosmo as oc
            import numpy as np
            collection = oc.open("haloproperties.hdf5", "haloparticles.hdf5")

            def computation(halo_properties, dm_particles):
                dx = np.mean(dm_particles.data["x"]) - halo_properties["fof_halo_center_x"]
                dy = np.mean(dm_particles.data["y"]) - halo_properties["fof_halo_center_y"]
                dz = np.mean(dm_particles.data["z"]) - halo_properties["fof_halo_center_z"]
                offset = np.sqrt(dx**2 + dy**2 + dz**2)
                return offset / halo_properties["sod_halo_radius"]

            collection = collection.evaluate(
                computation,
                name="offset",
                halo_properties=[
                    "fof_halo_center_x",
                    "fof_halo_center_y",
                    "fof_halo_center_z"
                    "sod_halo_radius"
                ],
                dm_particles=["x", "y", "z"]
            )

        The collection will now contain a column named "offset" with the results of the
        computation applied to each halo in the collection. Columns produced in this
        way will not respond to changes in unit convention.

        It is not required to pass a list of column names for a given dataset. If a list
        is not provided, all columns will be passed to the computation function. Data will
        be passed into the function as numpy arrays or astropy tables, depending on the
        value of the "format" argument. However if the evaluation involes a nested
        structure collection (e.g. a galaxy collection inside a structure collection)
        in addition to other datasets, the nested collection will be passed to your
        function as a StructureCollection.

        For more details and advanced usage see :ref:`Evaluating on Structure Collections`

        Parameters
        ----------

        func: Callable
            The function to evaluate on the rows in the dataset.

        dataset: Optional[str], default = None
            The dataset inside this collection to evaluate the function on. If none, assumes the function requires data from
            multiple datasets.

        vectorize: bool, default = False
            Whether to provide the values as full columns (True) or one row at a time (False) if evaluating on aa single dataset.
            Has no effect if evaluating over structures, since structures require input from multiple datasets which will not in
            general be the same length.

        insert: bool, default = True
            If true, the data will be inserted as a column in the specified dataset, or the main "properties" dataset
            if no dataset is specified. The new column will have the same name as the function. Otherwise the data
            will be returned directly.

        format: str, default = astropy
            Whether to provide data to your function as "astropy" quantities or "numpy" arrays/scalars. Default "astropy"

        **evaluate_kwargs: any,
            Any additional arguments that are required for your function to run. These will be passed directly
            to the function as keyword arguments. If a kwarg is an array of values with the same length as the dataset,
            it will be treated as an additional column.

        """
        if dataset is not None:
            datasets = dataset.split(".", 1)
            ds = self[datasets[0]]
            if isinstance(ds, oc.Dataset) and len(datasets) > 1:
                raise ValueError("Datasets cannot be nested!")
            elif isinstance(ds, oc.Dataset):
                result = ds.evaluate(
                    func,
                    format=format,
                    vectorize=vectorize,
                    insert=insert,
                    **evaluate_kwargs,
                )
            elif isinstance(ds, StructureCollection):
                ds_name = datasets[1] if len(datasets) > 1 else None
                result = ds.evaluate(
                    func,
                    ds_name,
                    format=format,
                    vectorize=vectorize,
                    insert=insert,
                    **evaluate_kwargs,
                )

            if result is None or not insert:
                return result

            assert isinstance(result, (oc.Dataset, StructureCollection))
            if ds.dtype == self.__source.dtype:
                new_source = result
                new_datasets = self.__datasets
            else:
                new_source = self.__source
                new_datasets = {**self.__datasets, datasets[0]: result}
            return StructureCollection(
                new_source,
                self.__header,
                new_datasets,
                self.__hide_source,
            )
        else:
            known_datasets = set(self.keys())
            kwarg_names = set(evaluate_kwargs.keys())

            requested_datasets = kwarg_names.intersection(known_datasets)
            other_kwarg_names = kwarg_names.difference(known_datasets)

            columns = {key: evaluate_kwargs[key] for key in requested_datasets}
            kwargs = {key: evaluate_kwargs[key] for key in other_kwarg_names}

            output = evaluate.visit_structure_collection(
                func, columns, self, format=format, evaluator_kwargs=kwargs
            )
            if not insert or output is None:
                return output
            return self.with_new_columns(**output, dataset=self.__source.dtype)

    def filter(self, *masks, on_galaxies: bool = False) -> StructureCollection:
        """
        Apply a filter to the halo or galaxy properties. Filters are constructed with
        :py:func:`opencosmo.col` and behave exactly as they would in
        :py:meth:`opencosmo.Dataset.filter`.

        If the collection contains both halos and galaxies, the filter can be applied to
        the galaxy properties dataset by setting `on_galaxies=True`. However this will
        filter for *halos* that host galaxies that match this filter. As a result,
        galxies that do not match this filter will remain if another galaxy in their
        host halo does match.

        See :ref:`Querying in Collections` for some examples.


        Parameters
        ----------
        *filters: Mask
            The filters to apply to the properties dataset constructed with
            :func:`opencosmo.col`.

        on_galaxies: bool, optional
            If True, the filter is applied to the galaxy properties dataset.

        Returns
        -------
        StructureCollection
            A new collection filtered by the given masks.

        Raises
        -------
        ValueError
            If on_galaxies is True but the collection does not contain
            a galaxy properties dataset.
        """
        if not masks:
            return self
        if not on_galaxies or self.__source.dtype == "galaxy_properties":
            filtered = self.__source.filter(*masks)
        elif "galaxy_properties" not in self.__datasets:
            raise ValueError("Dataset galaxy_properties not found in collection.")
        else:
            galaxy_properties = self["galaxy_properties"]
            assert isinstance(galaxy_properties, oc.Dataset)
            filtered = filter_source_by_dataset(
                galaxy_properties, self.__source, self.__header, *masks
            )
        return StructureCollection(
            filtered, self.__header, self.__datasets, self.__hide_source
        )

    def select(
        self, **column_selections: str | Iterable[str] | dict
    ) -> StructureCollection:
        """
        Update a dataset in the collection collection to only include the
        columns specified. The name of the arguments to this function should be
        dataset names. For example:

        .. code-block:: python

            collection = collection.select(
                halo_properties = ["fof_halo_mass", "sod_halo_mass", "sod_halo_cdelta"],
                dm_particles = ["x", "y", "z"]
            )

        Datasets that do not appear in the argument list will not be modified. You can
        remove entire datasets from the collection with
        :py:meth:`with_datasets <opencosmo.StructureCollection.with_datasets>`

        For nested structure collections, such as galaxies within halos, you can pass
        a nested dictionary:

        .. code-block:: python

            collection = oc.open("haloproperties.hdf5", "haloparticles.hdf5", "galaxyproperties.hdf5", "galaxyparticles.hdf5")

            collection = collection.select(
                halo_properties = ["fof_halo_mass", "sod_halo_mass", "sod_halo_cdelta"],
                dm_particles = ["x", "y", "z"]
                galaxies = {
                    "galaxy_properties": ["gal_mass_bar", "gal_mass_star"],
                    "star_particles": ["x", "y", "z"]
                }
            )


        Parameters
        ----------
        **column_selections : str | Iterable[str] | dict[str, Iterable[str]]
            The columns to select from a given dataset or sub-collection

        dataset : str
            The dataset to select from.

        Returns
        -------
        StructureCollection
            A new collection with only the selected columns for the specified dataset.

        Raises
        -------
        ValueError
            If the specified dataset is not found in the collection.
        """
        if not column_selections:
            return self
        new_source = self.__source
        new_datasets = {}
        for dataset, columns in column_selections.items():
            if dataset == self.__header.file.data_type:
                new_source = self.__source.select(columns)
                continue

            elif dataset not in self.__datasets:
                raise ValueError(f"Dataset {dataset} not found in collection.")

            new_ds = self.__datasets[dataset]

            if not isinstance(new_ds, oc.Dataset):
                if not isinstance(columns, dict):
                    raise ValueError(
                        "When working with nested structure collections, the argument should be a dictionary!"
                    )
                new_ds = new_ds.select(**columns)
            else:
                new_ds = new_ds.select(columns)

            new_datasets[dataset] = new_ds

        return StructureCollection(
            new_source,
            self.__header,
            self.__datasets | new_datasets,
            self.__hide_source,
        )

    def drop(self, **columns_to_drop):
        """
        Update the linked collection by dropping the specified columns
        in the specified datasets. This method follows the exact same semantics as
        :py:meth:`StructureCollection.select <opencosmo.StructureCollection.select>`.
        Argument names should be datasets in this collection, and the argument
        values should be a string, list of strings, or dictionary.

        Datasets that are not included will not be modified. You can drop
        entire datasets with :py:meth:`with_datasets <opencosmo.StructureCollection.with_datasets>`

        Parameters
        ----------
        **columns_to_drop : str | Iterable[str]
            The columns to drop from the dataset.

        dataset : str, optional
            The dataset to select from. If None, the properties dataset is used.

        Returns
        -------
        StructureCollection
            A new collection with only the selected columns for the specified dataset.

        Raises
        -------
        ValueError
            If the specified dataset is not found in the collection.
        """
        if not columns_to_drop:
            return self
        new_source = self.__source
        new_datasets = {}

        for dataset_name, columns in columns_to_drop.items():
            if dataset_name == self.__header.file.data_type:
                new_source = self.__source.drop(columns)
                continue

            elif dataset_name not in self.__datasets:
                raise ValueError(f"Dataset {dataset_name} not found in collection.")
            new_ds = self.__datasets[dataset_name]
            if isinstance(new_ds, oc.Dataset):
                new_ds = new_ds.drop(columns)
            elif isinstance(new_ds.StructureCollection):
                new_ds = new_ds.drop(**columns)

            new_datasets[dataset_name] = new_ds

        return StructureCollection(
            new_source,
            self.__header,
            self.__datasets | new_datasets,
            self.__hide_source,
        )

    def sort_by(self, column: str, invert: bool = False) -> StructureCollection:
        """
        Re-order the collection based on one of the structure collection's properties. Each
        StructureCollection contains a halo_properties or galaxy_properties dataset that
        contains the high-level measured properties of the structures in this collection.
        This method always operates on that dataset.

        Parameters
        ----------
        column : str
            The column in the halo_properties or galaxy_properties dataset to
            order the collection by.

        invert : bool, default = False
            If False (the default), ordering will be from least to greatest.
            Otherwise greatest to least.

        Returns
        -------
        result : StructureCollection
            A new StructureCollection ordered by the given column.

        """

        new_source = self.__source.sort_by(column, invert=invert)
        return StructureCollection(
            new_source,
            self.__header,
            self.__datasets,
            self.__hide_source,
        )

    def with_units(
        self,
        convention: Optional[str] = None,
        conversions: dict[u.Unit, u.Unit] = {},
        **dataset_conversions: dict,
    ):
        """
        Apply the given unit convention to the collection, or convert a subset
        of the columns in one or more of these datasets into a compatible
        unit.

        Because this collection contains several datasets, you must specify
        the dataset when performing conversions. For example, the equivalent
        unit conversion to the final one in the example in
        :py:meth:`opencosmo.Dataset.with_units` looks like this:

        .. code-block:: python

            import astropy.units as u

            structures = structures.with_units(
                "physical",
                halo_properties={"fof_halo_mass": u.kg, "fof_halo_center_x": u.ly}
            )

        You can use :code:`conversions` to specify a conversion that applies to all
        columns in the collection with the given unit, or specify per-dataset conversions.
        Per-dataset conversions always take precedent over collection-wide conversions.
        For example:

        .. code-block:: python

            import astropy.units as u

            conversions = {u.Mpc: u.lyr}
            structures = structures.with_units(
                conversions=conversions
                halo_properties = {
                    "conversions": {u.Mpc: u.km},
                    "fof_halo_center_x": u.m
                }
            )

        In this example, all values in Mpc will be converted to lightyears, except in the "halo_properties" dataset,
        where they will be converted to kilometers. The column "fof_halo_center_x" in "halo_properties" will
        be converted to meters instead.

        For more information, see :doc:`units`

        Parameters
        ----------
        convention : str
            The unit convention to apply. One of "unitless", "scalefree",
            "comoving", or "physical".

        conversions : dict[astropy.units.Unit, astropy.units.Unit]
            Unit conversions to apply across all columns in the collection

        **dataset_conversion : dict
            Unit conversions apply to specific datasets in the collection.

        Returns
        -------
        StructureCollection
            A new collection with the unit convention applied.
        """
        if conversions:
            for ds_name in self.keys():
                ds_conversions = dataset_conversions.get(ds_name, {})
                new_ds_conversions = conversions | ds_conversions.get("conversions", {})
                ds_conversions["conversions"] = new_ds_conversions
                dataset_conversions[ds_name] = ds_conversions

        conversion_keys = set(dataset_conversions.keys())
        unknown = conversion_keys.difference(self.keys())
        if unknown:
            raise ValueError(f"Unknown datasets in conversions: {unknown}")

        if self.__source.dtype in conversion_keys or (
            not conversion_keys and convention is None
        ):
            new_source = self.__source.with_units(
                convention, **dataset_conversions.get(self.__source.dtype, {})
            )
        else:
            new_source = self.__source
        new_datasets = {}
        for key, dataset in self.__datasets.items():
            ds_conversions = dataset_conversions.get(key, {})
            if convention is None and not ds_conversions:
                new_datasets[key] = dataset.with_units()
                continue
            new_ds = dataset.with_units(convention, **ds_conversions)
            new_datasets[key] = new_ds

        return StructureCollection(
            new_source, self.__header, new_datasets, self.__hide_source
        )

    def take(self, n: int, at: str = "random"):
        """
        Take some number of structures from the collection.
        See :py:meth:`opencosmo.Dataset.take`.

        Parameters
        ----------
        n : int
            The number of structures to take from the collection.
        at : str, optional
            The method to use to take the structures. One of "random", "first",
            or "last". Default is "random".

        Returns
        -------
        StructureCollection
            A new collection with the structures taken from the original.
        """
        new_source = self.__source.take(n, at)
        return StructureCollection(
            new_source,
            self.__header,
            self.__datasets,
            self.__hide_source,
        )

    def take_range(self, start: int, end: int):
        """
        Create a new collection from a row range in this collection. We use standard
        indexing conventions, so the rows included will be start -> end - 1.

        Parameters
        ----------
        start : int
            The first row to get.
        end : int
            The last row to get.

        Returns
        -------
        table : astropy.table.Table
            The table with only the rows from start to end.

        Raises
        ------
        ValueError
            If start or end are negative or greater than the length of the dataset
            or if end is greater than start.

        """
        new_source = self.__source.take_range(start, end)
        return StructureCollection(
            new_source, self.__header, self.__datasets, self.__hide_source
        )

    def take_rows(self, rows: np.ndarray | DataIndex):
        """
        Take the rows of this collection  specified by the :code:`rows` argument.
        :code:`rows` should be an array of integers.

        Parameters:
        -----------
        rows: np.ndarray[int]

        Returns
        -------
        dataset: The dataset with only the specified rows included

        Raises:
        -------
        ValueError:
            If any of the indices is less than 0 or greater than the length of the
            dataset.

        """
        new_source = self.__source.take_rows(rows)
        return StructureCollection(
            new_source, self.__header, self.__datasets, self.__hide_source
        )

    def with_new_columns(
        self,
        dataset: str,
        descriptions: str | dict[str, str] = {},
        **new_columns: DerivedColumn,
    ):
        """
        Add new column(s) to one of the datasets in this collection. This behaves
        exactly like :py:meth:`oc.Dataset.with_new_columns`, except that you must
        specify which dataset the columns should refer too.

        .. code-block:: python

            pe = oc.col("phi") * oc.col("mass")
            collection = collection.with_new_columns("dm_particles", pe=pe)

        Structure collections can hold other structure collections. For example, a
        collection of Halos may hold a structure collection that contians the galaxies
        of those halos. To update datasets within these collections, use dot syntax
        to specify a path:

        .. code-block:: python

            pe = oc.col("phi") * oc.col("mass")
            collection = collection.with_new_columns("galaxies.star_particles", pe=pe)

        You can also pass numpy arrays or astropy quantities:

        .. code-block:: python

            random_value = np.random.randint(0, 90, size=len(collection))
            random_quantity = random_value*u.deg

            collection = collection.with_new_columns("halo_properties",
                random_quantity=random_quantity)

        See :ref:`Adding Custom Columns` for more examples.


        Parameters
        ----------
        dataset : str
            The name of the dataset to add columns to

        descriptions : str | dict[str, str], optional
            Descriptions for the new columns. These descriptions will be accessible through
            :py:attr:`Dataset.descriptions <opencosmo.Dataset.descriptions>`. If a dictionary,
            should have keys matching the column names.

        ** columns: opencosmo.DerivedColumn
            The new columns

        Returns
        -------
        new_collection : opencosmo.StructureCollection
            This collection with the additional columns added

        Raise
        -----
        ValueError
            If the dataset is not found in this collection
        """
        path = dataset.split(".")
        if len(path) > 1:
            collection_name = path[0]
            if collection_name not in self.keys():
                raise ValueError(f"No collection {collection_name} found!")
            new_collection = self.__datasets[collection_name]
            if not isinstance(new_collection, StructureCollection):
                raise ValueError(f"{collection_name} is not a collection!")
            new_collection = new_collection.with_new_columns(
                ".".join(path[1:]), descriptions=descriptions, **new_columns
            )
            return StructureCollection(
                self.__source,
                self.__header,
                {**self.__datasets, collection_name: new_collection},
                self.__hide_source,
            )

        if dataset == self.__source.dtype:
            new_source = self.__source.with_new_columns(
                **new_columns, descriptions=descriptions
            )
            return StructureCollection(
                new_source,
                self.__header,
                self.__datasets,
            )
        elif dataset not in self.__datasets.keys():
            raise ValueError(f"Dataset {dataset} not found in this collection!")

        ds = self.__datasets[dataset]

        if not isinstance(ds, oc.Dataset):
            raise ValueError(f"{dataset} is not a dataset!")

        new_ds = ds.with_new_columns(**new_columns, descriptions=descriptions)
        return StructureCollection(
            self.__source,
            self.__header,
            {**self.__datasets, dataset: new_ds},
            self.__hide_source,
        )

    def objects(
        self, data_types: Optional[Iterable[str]] = None, ignore_empty=True
    ) -> Iterable[dict[str, Any]]:
        """
        Iterate over the objects in this collection as pairs of
        (properties, datasets). For example, a halo collection could yield
        the halo properties and datasets for each of the associated partcles.

        If you don't need all the datasets, you can specify a list of data types
        for example:

        .. code-block:: python

            for row, particles in
                collection.objects(data_types=["gas_particles", "star_particles"]):
                # do work

        At each iteration, "row" will be a dictionary of halo properties with associated
        units, and "particles" will be a dictionary of datasets with the same keys as
        the data types.
        """
        if data_types is None:
            data_types = self.__datasets.keys()

        data_types = list(data_types)
        if not all(dt in self.__datasets for dt in data_types):
            raise ValueError("Some data types are not linked in the collection.")

        if len(self) == 0:
            warn("Tried to iterate over a collection with no structures in it!")
            return

        link_handler = None
        data_columns = set(self.__source.columns)
        metadata_columns: list[str] = reduce(
            lambda acc, key: acc + self.__handler.columns[key], data_types, []
        )
        rename_galaxies = "galaxies" in self.keys()
        for row in self.__source.rows(metadata_columns=metadata_columns):
            row = dict(row)
            links = self.__handler.parse(row)
            output = {
                key: self.__datasets[key].take_rows(index)
                for key, index in links.items()
                if key in data_types
            }
            if not output:
                continue
            if not self.__hide_source:
                output.update({self.__source.dtype: row})
            yield output

    def with_datasets(self, datasets: list[str]):
        """
        Create a new collection out of a subset of the datasets in this collection.
        It is also possible to do this when you iterate over the collection with
        :py:meth:`StructureCollection.objects <opencosmo.StructureCollection.objects>`,
        however doing it up front may be more desirable if you don't plan to use
        the dropped datasets at any point.
        """

        if not isinstance(datasets, list):
            raise ValueError("Expected a list with at least one entry")

        known_datasets = set(self.keys())
        requested_datasets = set(datasets)
        if not requested_datasets.issubset(known_datasets):
            raise ValueError(f"Unknown datasets {requested_datasets - known_datasets}")

        if self.__source.dtype not in requested_datasets:
            hide_source = True
        else:
            hide_source = False
            requested_datasets.remove(self.__source.dtype)

        new_datasets = {name: self.__datasets[name] for name in requested_datasets}
        return StructureCollection(
            self.__source, self.__header, new_datasets, hide_source
        )

    def halos(self, *args, **kwargs):
        """
        Alias for "objects" in the case that this StructureCollection contains halos.
        """
        if self.__source.dtype == "halo_properties":
            yield from self.objects(*args, **kwargs)
        else:
            raise AttributeError("This collection does not contain halos!")

    def galaxies(self, *args, **kwargs):
        """
        Alias for "objects" in the case that this StructureCollection contains galaxies
        """
        if self.__source.dtype == "galaxy_properties":
            yield from self.objects(*args, **kwargs)
        else:
            raise AttributeError("This collection does not contain galaxies!")

    def make_schema(self) -> StructCollectionSchema:
        schema = StructCollectionSchema()
        source_name = self.__source.dtype

        for name, dataset in self.items():
            ds_schema = dataset.make_schema()
            if name == "galaxies":
                name = "galaxy_properties"
            schema.add_child(ds_schema, name)

        return schema



from __future__ import annotations

from typing import TYPE_CHECKING

from opencosmo.index import SimpleIndex
from opencosmo.spatial.region import HealPixRegion

if TYPE_CHECKING:
    from opencosmo.spatial.protocols import Region


class HealPixIndex:
    subdivision_factor = 4

    def __init__(self):
        pass

    def get_partition_region(self, index: SimpleIndex, level: int) -> Region:
        idxs = index.into_array()
        return HealPixRegion(idxs, 2**level)

    def query(
        self, region: Region, level: int = 1
    ) -> dict[int, tuple[SimpleIndex, SimpleIndex]]:
        """
        Raw healpix data is

        - pi < phi < pi
        0 < theta < pi

        SkyCoordinates are typically

        0 < RA < 360 deg
        - 90 deg < Dec < 90 deg

        And HealPix is

        0 < phi < 2*pi
        0 < theta < pi

        This is why we can't have nice things
        """
        if not hasattr(region, "get_healpix_intersections"):
            raise ValueError("Didn't recieve a 2D region!")
        nside = 2**level
        intersects = region.get_healpix_intersections(nside)
        return {level: (SimpleIndex.empty(), SimpleIndex(intersects))}



from __future__ import annotations

from functools import reduce
from itertools import cycle
from typing import TYPE_CHECKING, Iterable, Optional
from weakref import finalize

import astropy.units as u
import numpy as np
from astropy.table import QTable

from opencosmo.dataset.handler import Hdf5Handler
from opencosmo.dataset.im import InMemoryColumnHandler
from opencosmo.index import ChunkedIndex, SimpleIndex
from opencosmo.io import schemas as ios
from opencosmo.units import UnitConvention
from opencosmo.units.handler import make_unit_handler

if TYPE_CHECKING:
    import h5py
    from astropy import table, units
    from astropy.cosmology import Cosmology
    from numpy.typing import NDArray

    from opencosmo.dataset.column import DerivedColumn
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.index import DataIndex
    from opencosmo.spatial.protocols import Region
    from opencosmo.units.handler import UnitHandler


def deregister_state(id: int, handler: Hdf5Handler):
    handler.deregister(id)


class DatasetState:
    """
    Holds mutable state required by the dataset. Cleans up the dataset to mostly focus
    on very high-level operations. Not a user facing class.
    """

    def __init__(
        self,
        raw_data_handler: Hdf5Handler,
        im_handler: InMemoryColumnHandler,
        derived_columns: dict[str, DerivedColumn],
        unit_handler: UnitHandler,
        header: OpenCosmoHeader,
        columns: set[str],
        region: Region,
        sort_by: Optional[tuple[str, bool]],
    ):
        self.__raw_data_handler = raw_data_handler
        self.__im_handler = im_handler
        self.__derived_columns = derived_columns
        self.__unit_handler = unit_handler
        self.__header = header
        self.__columns = columns
        self.__region = region
        self.__sort_by = sort_by
        self.__raw_data_handler.register(self, self.__derived_columns)
        finalize(self, deregister_state, id(self), self.__raw_data_handler)

    def __rebuild(self, **updates):
        new = {
            "raw_data_handler": self.__raw_data_handler,
            "im_handler": self.__im_handler,
            "derived_columns": self.__derived_columns,
            "unit_handler": self.__unit_handler,
            "header": self.__header,
            "columns": self.__columns,
            "region": self.__region,
            "sort_by": self.__sort_by,
        } | updates
        return DatasetState(**new)

    def __exit__(self, *exec_details):
        return None

    @classmethod
    def from_group(
        cls,
        group: h5py.Group,
        header: OpenCosmoHeader,
        unit_convention: UnitConvention,
        region: Region,
        index: Optional[DataIndex] = None,
        metadata_group: Optional[h5py.Group] = None,
    ):
        handler = Hdf5Handler.from_group(group, index, metadata_group)
        unit_handler = make_unit_handler(handler.data, header, unit_convention)

        columns = set(handler.columns)
        im_handler = InMemoryColumnHandler.empty()
        return DatasetState(
            handler,
            im_handler,
            {},
            unit_handler,
            header,
            columns,
            region,
            None,
        )

    def __len__(self):
        return len(self.__raw_data_handler.index)

    @property
    def descriptions(self):
        return (
            self.__im_handler.descriptions
            | {name: col.description for name, col in self.__derived_columns.items()}
            | self.__raw_data_handler.descriptions
        )

    @property
    def raw_index(self):
        return self.__raw_data_handler.index

    @property
    def unit_handler(self):
        return self.__unit_handler

    @property
    def convention(self):
        return self.__unit_handler.current_convention

    @property
    def region(self):
        return self.__region

    @property
    def header(self):
        return self.__header

    @property
    def columns(self) -> list[str]:
        return list(self.__columns)

    @property
    def meta_columns(self) -> list[str]:
        return self.__raw_data_handler.metadata_columns

    def get_data(
        self,
        ignore_sort: bool = False,
        metadata_columns: list = [],
        unit_kwargs: dict = {},
    ) -> table.QTable:
        """
        Get the data for a given handler.
        """
        data = self.__build_derived_columns(unit_kwargs)
        data = data | self.__get_im_columns(data, unit_kwargs)

        raw_columns = set(self.columns).difference(data.keys())
        if (
            self.__sort_by is not None
            and self.__sort_by[0] in self.__raw_data_handler.columns
        ):
            raw_columns.add(self.__sort_by[0])

        raw_data = self.__raw_data_handler.get_data(raw_columns)
        data = data | self.__unit_handler.apply_units(raw_data, unit_kwargs)

        output = QTable(data, copy=False)

        data_columns = set(output.columns)
        raw_index_array = self.__raw_data_handler.index.into_array()

        if metadata_columns:
            output.update(self.__raw_data_handler.get_metadata(metadata_columns))

        if not ignore_sort and self.__sort_by is not None:
            order = output.argsort(self.__sort_by[0], reverse=self.__sort_by[1])
            output = output[order]
            raw_index_array = raw_index_array[order]

        extras = set(data.keys()).difference(self.columns)
        if extras:
            output.remove_columns(extras)

        return output

    def get_metadata(self, columns=[]):
        return self.__raw_data_handler.get_metadata(columns)

    def with_mask(self, mask: NDArray[np.bool_]):
        new_raw_handler = self.__raw_data_handler.mask(mask)
        new_im_handler = self.__im_handler.get_rows(mask)
        return self.__rebuild(
            im_handler=new_im_handler,
            raw_data_handler=new_raw_handler,
        )

    def make_schema(self):
        header = self.__header.with_region(self.__region)
        raw_columns = self.__columns.intersection(self.__raw_data_handler.columns)

        schema = self.__raw_data_handler.make_schema(raw_columns, header)
        derived_names = set(self.__derived_columns.keys()).intersection(self.columns)
        derived_data = (
            self.select(derived_names)
            .with_units("unitless", {}, {}, None, None)
            .get_data()
        )
        column_units = {
            name: self.__unit_handler.base_units[name] for name in self.columns
        }

        for colname in derived_names:
            attrs = {
                "unit": str(column_units[colname]),
                "description": self.__derived_columns[colname].description,
            }
            coldata = derived_data[colname].value
            colschema = ios.ColumnSchema(
                colname, ChunkedIndex.from_size(len(coldata)), coldata, attrs
            )
            schema.add_child(colschema, f"data/{colname}")

        im_names = set(self.__im_handler.keys()).intersection(self.columns)
        im_descriptions = self.__im_handler.descriptions
        for colname, coldata in self.__im_handler.columns():
            if colname not in im_names:
                continue
            attrs = {}
            attrs["unit"] = str(column_units.get(colname))
            attrs["description"] = im_descriptions.get(colname)

            colschema = ios.ColumnSchema(
                colname, ChunkedIndex.from_size(len(coldata)), coldata, attrs
            )
            schema.add_child(colschema, f"data/{colname}")

        return schema

    def with_new_columns(
        self,
        descriptions: dict[str, str] = {},
        **new_columns: DerivedColumn | np.ndarray | units.Quantity,
    ):
        """
        Add a set of derived columns to the dataset. A derived column is a column that
        has been created based on the values in another column.
        """
        derived_update = {}
        new_unit_handler = self.__unit_handler
        new_im_handler = self.__im_handler
        for name, new_col in new_columns.items():
            if name in self.__columns:
                raise ValueError(f"Dataset already has column named {name}")

            if isinstance(new_col, np.ndarray):
                if len(new_col) != len(self):
                    raise ValueError(
                        f"New column {name} has length {len(new_col)} but this dataset "
                        "has length {len(self.__index)}"
                    )
                if isinstance(new_col, u.Quantity):
                    new_unit_handler = new_unit_handler.with_static_columns(
                        **{name: new_col.unit}
                    )
                    new_col = new_col.value
                else:
                    new_unit_handler = new_unit_handler.with_new_columns(**{name: None})

                new_im_handler = new_im_handler.with_new_column(
                    name, new_col, descriptions.get(name, "None")
                )

            elif not new_col.check_parent_existance(self.__columns):
                raise ValueError(
                    f"Column {name} is derived from columns "
                    "that are not in the dataset!"
                )
            else:
                unit = new_col.get_units(self.__unit_handler.base_units)
                new_col.description = descriptions.get(name, "None")
                new_unit_handler = new_unit_handler.with_new_columns(**{name: unit})
                derived_update[name] = new_col

        new_derived = self.__derived_columns | derived_update
        new_columns_ = self.__columns.union(new_columns.keys())
        return self.__rebuild(
            unit_handler=new_unit_handler,
            im_handler=new_im_handler,
            derived_columns=new_derived,
            columns=new_columns_,
        )

    def __build_derived_columns(self, unit_kwargs: dict) -> table.Table:
        """
        Build any derived columns that are present in this dataset
        """
        if not self.__derived_columns:
            return {}

        derived_names = set(self.__derived_columns.keys()).intersection(self.columns)
        if (
            self.__sort_by is not None
            and self.__sort_by[0] in self.__derived_columns.keys()
        ):
            derived_names.add(self.__sort_by[0])

        ancestors: set[str] = reduce(
            lambda acc, der: acc.union(der.requires()),
            self.__derived_columns.values(),
            set(),
        )
        raw_ancestors = ancestors.intersection(self.__raw_data_handler.columns)
        im_ancestors = ancestors.intersection(self.__im_handler.keys())
        additional_derived = ancestors.intersection(self.__derived_columns.keys())
        derived_names = derived_names.union(additional_derived)

        data = self.__raw_data_handler.get_data(raw_ancestors)
        data = data | self.__im_handler.get_data(im_ancestors)
        data = self.__unit_handler.apply_units(data, unit_kwargs)
        seen: set[str] = set()
        for name in cycle(derived_names):
            if derived_names.issubset(data.keys()):
                break
            elif name in seen:
                # We're stuck in a loop
                raise ValueError(
                    "Something went wrong when trying to instatiate derived columns!"
                )
            elif name in data:
                continue
            elif set(data.keys()).issuperset(self.__derived_columns[name].requires()):
                data[name] = self.__derived_columns[name].evaluate(data)
                seen = set()
            else:
                seen.add(name)

        return data

    def __get_im_columns(self, data: dict, unit_kwargs) -> table.Table:
        im_data = {}
        for colname, column in self.__im_handler.columns():
            im_data[colname] = column

        return self.__unit_handler.apply_units(im_data, unit_kwargs)

    def with_region(self, region: Region):
        """
        Return the same dataset but with a different region
        """
        return self.__rebuild(region=region)

    def select(self, columns: str | Iterable[str]):
        """
        Select a subset of columns from the dataset. It is possible for a user to select
        a derived column in the dataset, but not the columns it is derived from.
        This class tracks any columns which are required to materialize the dataset but
        are not in the final selection in self.__hidden. When the dataset is
        materialized, the columns in self.__hidden are removed before the data is
        returned to the user.

        """
        if isinstance(columns, str):
            columns = [columns]

        columns = set(columns)
        missing = columns - self.__columns
        if missing:
            raise ValueError(
                f"Tried to select columns that are not in this dataset: {missing}"
            )

        return self.__rebuild(columns=columns)

    def sort_by(self, column_name: str, invert: bool):
        if column_name not in self.columns:
            raise ValueError(f"This dataset has no column {column_name}")

        return self.__rebuild(sort_by=(column_name, invert))

    def get_sorted_index(self):
        if self.__sort_by is not None:
            column = self.select(self.__sort_by[0]).get_data(ignore_sort=True)[
                self.__sort_by[0]
            ]
            sorted = np.argsort(column)
            if self.__sort_by[1]:
                sorted = sorted[::-1]

        else:
            sorted = None

        return sorted

    def take(self, n: int, at: str):
        """
        Take rows from the dataset.
        """

        sorted = self.get_sorted_index()
        take_index: DataIndex

        if at == "start":
            return self.take_range(0, n)
        elif at == "end":
            return self.take_range(len(self) - n, len(self))
        elif at == "random":
            row_indices = np.random.choice(len(self), n, replace=False)
            take_index = SimpleIndex(row_indices)

        new_handler = self.__raw_data_handler.take(take_index, sorted)
        new_im_handler = self.__im_handler.take(take_index, sorted)

        return self.__rebuild(
            raw_data_handler=new_handler,
            im_handler=new_im_handler,
        )

    def take_range(self, start: int, end: int):
        """
        Take a range of rows form the dataset.
        """
        if start < 0 or end < 0:
            raise ValueError("start and end must be positive.")
        if end < start:
            raise ValueError("end must be greater than start.")
        if end > len(self.__raw_data_handler.index):
            raise ValueError("end must be less than the length of the dataset.")

        if start < 0 or end > len(self.__raw_data_handler.index):
            raise ValueError("start and end must be within the bounds of the dataset.")

        sorted = self.get_sorted_index()

        take_index = ChunkedIndex.single_chunk(start, end - start)

        new_raw_handler = self.__raw_data_handler.take(take_index, sorted)
        new_im = self.__im_handler.take(take_index, sorted)
        return self.__rebuild(
            raw_data_handler=new_raw_handler,
            im_handler=new_im,
        )

    def take_rows(self, rows: DataIndex):
        if len(self) == 0:
            return self
        if rows.range()[1] > len(self) or rows.range()[0] < 0:
            raise ValueError(
                f"Row indices must be between 0 and the length of this dataset!"
            )
        sorted = self.get_sorted_index()
        new_handler = self.__raw_data_handler.take(rows, sorted)
        new_im_handler = self.__im_handler.take(rows, sorted)

        return self.__rebuild(
            raw_data_handler=new_handler,
            im_handler=new_im_handler,
        )

    def with_units(
        self,
        convention: Optional[str],
        conversions: dict[u.Unit, u.Unit],
        columns: dict[str, u.Unit],
        cosmology: Cosmology,
        redshift: float | table.Column,
    ):
        """
        Change the unit convention
        """

        if convention is None:
            convention_ = self.__unit_handler.current_convention
        else:
            convention_ = UnitConvention(convention)
        if (
            convention_ == UnitConvention.SCALEFREE
            and UnitConvention(self.header.file.unit_convention)
            != UnitConvention.SCALEFREE
        ):
            raise ValueError(
                f"Cannot convert units with convention {self.header.file.unit_convention} to convention scalefree"
            )
        column_keys = set(columns.keys())
        missing_columns = column_keys - set(self.columns)
        if missing_columns:
            raise ValueError(f"Dataset does not have columns {missing_columns}")

        new_handler = self.__unit_handler.with_convention(convention_).with_conversions(
            conversions, columns
        )
        return self.__rebuild(unit_handler=new_handler)



from .cosmology import CosmologyParameters
from .file import FileParameters
from .hacc import HaccSimulationParameters
from .parameters import read_header_attributes, write_header_attributes

__all__ = [
    "FileParameters",
    "read_header_attributes",
    "write_header_attributes",
    "CosmologyParameters",
    "HaccSimulationParameters",
]



from .io import get_collection_type, open_simulation_files
from .lightcone import Lightcone
from .protocols import Collection
from .simulation import SimulationCollection
from .structure import StructureCollection, open_linked_files

__all__ = [
    "Collection",
    "SimulationCollection",
    "StructureCollection",
    "open_multi_dataset_file",
    "SimulationCollection",
    "open_simulation_files",
    "open_linked_files",
    "Lightcone",
    "get_collection_type",
]



from __future__ import annotations

from copy import copy
from functools import cache
from itertools import chain
from types import UnionType
from typing import TYPE_CHECKING, Any, Optional

import h5py
from pydantic import BaseModel, ValidationError

from opencosmo.file import broadcast_read, file_reader, file_writer
from opencosmo.parameters import (
    FileParameters,
    dtype,
    origin,
    read_header_attributes,
    write_header_attributes,
)
from opencosmo.parameters.units import apply_units
from opencosmo.units import UnitConvention

if TYPE_CHECKING:
    from pathlib import Path


class OpenCosmoHeader:
    """
    A class to represent the header of an OpenCosmo file. The header contains
    information about the simulation the data is a part of, as well as other
    meatadata that are useful to the library in various contexts. Most files
    will have a single unique header, but it is possible to have multiple
    headers in a SimulationCollection.
    """

    def __init__(
        self,
        file_pars: FileParameters,
        required_origin_parameters: dict[str, BaseModel],
        optional_origin_parameters: dict[str, BaseModel],
        dtype_parameters: dict[str, BaseModel],
        unit_convention: UnitConvention = UnitConvention.SCALEFREE,
    ):
        self.__file_pars = file_pars
        self.__required_origin_parameters = required_origin_parameters
        self.__optional_origin_parameters = optional_origin_parameters
        self.__dtype_parameters = dtype_parameters
        self.unit_convention = unit_convention

    def __eq__(self, other):
        return (
            self.__file_pars == other.__file_pars
            and self.__required_origin_parameters == other.__required_origin_parameters
            and self.__optional_origin_parameters == other.__optional_origin_parameters
            and self.__dtype_parameters == other.__dtype_parameters
        )

    def __hash__(self):
        # Create a frozenset of the items in the dictionary
        # Each item is a tuple of (key, hash of the model)
        iter_ = chain(
            {"file": self.__file_pars}.items(),
            {"unit_convention": self.unit_convention}.items(),
            self.__required_origin_parameters.items(),
            self.__optional_origin_parameters.items(),
            self.__dtype_parameters.items(),
        )
        s = frozenset((key, hash(model)) for key, model in iter_)
        return hash(s)

    def with_units(self, convention: UnitConvention | str):
        convention = UnitConvention(convention)
        if convention == self.unit_convention:
            return self
        return OpenCosmoHeader(
            self.__file_pars,
            self.__required_origin_parameters,
            self.__optional_origin_parameters,
            self.__dtype_parameters,
            convention,
        )

    @cache
    def __get_access_table(self):
        table = {}
        all_models = chain(
            self.__required_origin_parameters.values(),
            self.__optional_origin_parameters.values(),
            self.__dtype_parameters.values(),
        )
        for model in all_models:
            if not hasattr(model, "ACCESS_PATH"):
                continue
            table[model.ACCESS_PATH] = model
            if hasattr(model, "ACCESS_TRANSFORMATION"):
                table[model.ACCESS_PATH] = model.ACCESS_TRANSFORMATION()

        cosmology = table.get("cosmology")
        convention = object.__getattribute__(self, "unit_convention")
        scale_factor = cosmology.scale_factor(self.__file_pars.redshift)
        for name, model in table.items():
            if isinstance(model, BaseModel):
                table[name] = apply_units(
                    model,
                    cosmology,
                    convention,
                    unit_kwargs={"scale_factor": scale_factor},
                )

        return table

    @property
    def parameters(self):
        """
        Return the parametrs stored in this header as
        key-value pairs. The values will be Pydantic models.

        Any block of parameters that is returned from this method can
        also be accessed with standard dot notation. For example, HACC
        data contains a "simulation" block that contains the parameters that
        were used to run the original simulation. The following calls are
        equivalent:

        .. code-block:: python

            header.simulation
            header.parameters["simulation"]

        Returns
        -------
        parameters: dict[str, pydantic.BaseModel]
            The parameter blocks associated with this header

        """
        return self.__get_access_table()

    def __getattr__(self, key: str):
        if key.startswith("__"):  # avoid infinite recursion when serailizing for MPI
            raise AttributeError(key)

        table = self.__get_access_table()
        try:
            return table[key]
        except KeyError:
            return object.__getattribute__(self, key)

    def with_region(self, region):
        if region is not None:
            region_model = region.into_model()
        else:
            region_model = None
        new_file_pars = self.__file_pars.model_copy(update={"region": region_model})
        new_header = OpenCosmoHeader(
            new_file_pars,
            self.__required_origin_parameters,
            self.__optional_origin_parameters,
            self.__dtype_parameters,
        )
        return new_header

    def with_parameter(self, key: str, value: Any):
        """
        Update a dtype parameter with a new value. This in general should never
        be called by the user. Returns a copy.
        """
        path = key.split("/")
        if len(path) != 2:
            raise ValueError("Can only update top-level dtype parameters")
        new_dtype_parameters = copy(self.__dtype_parameters)
        model = new_dtype_parameters[path[0]]
        new_model = model.model_copy(update={path[1]: value})
        new_dtype_parameters[path[0]] = new_model
        return OpenCosmoHeader(
            self.__file_pars,
            self.__required_origin_parameters,
            self.__optional_origin_parameters,
            new_dtype_parameters,
        )

    def with_parameters(self, updates: dict[str, Any]):
        if not updates:
            return self
        new_header = self
        for key, val in updates.items():
            new_header = new_header.with_parameter(key, val)
        return new_header

    def write(self, file: h5py.File | h5py.Group) -> None:
        write_header_attributes(file, "file", self.__file_pars)
        to_write = chain(
            self.__required_origin_parameters.items(),
            self.__optional_origin_parameters.items(),
            self.__dtype_parameters.items(),
        )
        for path, data in to_write:
            write_header_attributes(file, path, data)

    @property
    def file(self) -> FileParameters:
        """
        All files must at minimum have a "file" block in their header. This block
        contains basic information like the original source of the data and
        its data type.
        """
        return self.__file_pars


@file_writer
def write_header(
    path: Path, header: OpenCosmoHeader, dataset_name: Optional[str] = None
) -> None:
    """
    Write the header of an OpenCosmo file

    Parameters
    ----------
    file : h5py.File
        The file to write to
    header : OpenCosmoHeader
        The header information to write

    """
    with h5py.File(path, "w") as f:
        if dataset_name is not None:
            group = f.require_group(dataset_name)
        else:
            group = f
        header.write(group)


@broadcast_read
@file_reader
def read_header(
    file: h5py.File | h5py.Group,
    unit_convention: UnitConvention = UnitConvention.COMOVING,
) -> OpenCosmoHeader:
    """
    Read the header of an OpenCosmo file

    This function may be useful if you just want to access some basic
    information about the simulation but you don't plan to actually
    read any data.

    Parameters
    ----------
    file : str | Path
        The path to the file

    Returns
    -------
    header : OpenCosmoHeader
        The header information from the file


    """
    try:
        file_parameters = read_header_attributes(file, "file", FileParameters)
    except KeyError as e:
        raise KeyError(
            "File header is malformed. Are you sure it is an OpenCosmo file?\n "
            f"Error: {e}"
        )

    origin_parameter_models = origin.get_origin_parameters(file_parameters.origin)
    required_origin_params, optional_origin_params = read_origin_parameters(
        file, origin_parameter_models
    )
    dtype_parameter_models = dtype.get_dtype_parameters(file_parameters)
    dtype_params = read_dtype_parameters(file, dtype_parameter_models)

    h = OpenCosmoHeader(
        file_parameters,
        required_origin_params,
        optional_origin_params,
        dtype_params,
        unit_convention,
    )
    return h


def read_origin_parameters(
    file: h5py.File | h5py.Group,
    origin_parameters: dict[str, dict[str, type[BaseModel]]],
):
    """
    An "origin" describes the original source of a given dataset. Currently the only
    origin we support in the OpenCosmo toolkit is HACC.

    Origins can define a set of required and optional parameters.
    """
    required = origin_parameters["required"]
    required_output = {}
    for path, model in required.items():
        if isinstance(model, UnionType):
            required_output[path] = load_union_model(file, path, model)
        else:
            required_output[path] = read_header_attributes(file, path, model)

    optional_output = {}
    optional = origin_parameters["optional"]
    for path, model in optional.items():
        if isinstance(origin, UnionType):
            read_fn = load_union_model
        else:
            read_fn = read_header_attributes
        try:
            optional_output[path] = read_fn(file, path, model)
        except (ValidationError, KeyError):
            continue

    return required_output, optional_output


def read_dtype_parameters(
    file: h5py.File | h5py.Group, dtype_paramter_models: dict[str, type[BaseModel]]
):
    """
    Data types can also define parameters that they expect. For now, all dtype
    parameters are required. They MUST define an "ACCESS_PATH" attribute,
    which tells the header how users should be allowed to access them.

    """
    dtype_output = {}
    for path, model in dtype_paramter_models.items():
        if isinstance(model, UnionType):
            dtype_output[path] = load_union_model(file, path, model)
        else:
            dtype_output[path] = read_header_attributes(file, path, model)

        if not hasattr(model, "ACCESS_PATH"):
            # This should be always always always caught in testing
            raise ValueError(f"Model {model} does not have an access path!")

    return dtype_output


def load_union_model(
    file: h5py.File | h5py.Group, path: str, allowed_models: UnionType, **kwargs
):
    for model in allowed_models.__args__:
        try:
            return read_header_attributes(file, path, model)
        except ValidationError as ve:
            if any(e["type"] == "missing" or e["input"] is None for e in ve.errors()):
                continue
            else:
                raise ValueError(
                    f"Parsing header paramter model raised a validation error: \n {ve}"
                )
    raise ValueError("Input attributes do not match any of the models in the union")



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Type

from astropy import cosmology  # type: ignore

from opencosmo import header
from opencosmo.file import broadcast_read, file_reader

if TYPE_CHECKING:
    import h5py

    from opencosmo.parameters import CosmologyParameters

"""
Reads cosmology from the header of the file and returns the
astropy.cosmology object.
"""


@broadcast_read
@file_reader
def read_cosmology(file: h5py.File) -> cosmology.Cosmology:
    """
    Read cosmology from the header of an OpenCosmo file

    This function reads the cosmology parameters from the
    header of an OpenCosmo file and returns the most specific
    astropy.Cosmology object that it can. For example, it can
    distinguish between FlatLambdaCDM and non-flat wCDM models.

    Parameters
    ----------
    file : str | Path
        The path to the file

    Returns
    -------
    cosmology : astropy.Cosmology
        The cosmology object corresponding to the cosmology in the file
    """

    head = header.read_header(file)
    # The header reads parameters and calls into the code
    # below to produce an actual cosmology object.
    return head.cosmology


def make_cosmology(parameters: "CosmologyParameters") -> cosmology.Cosmology:
    cosmology_type = get_cosmology_type(parameters)
    expected_arguments = inspect.signature(cosmology_type).parameters.keys()
    input_paremeters = {}
    for argname in expected_arguments:
        try:
            input_paremeters[argname] = getattr(parameters, argname)
        except AttributeError:
            continue
    return cosmology_type(**input_paremeters)


def get_cosmology_type(parameters: "CosmologyParameters") -> Type[cosmology.Cosmology]:
    is_flat = (parameters.Om0 + parameters.Ode0) == 1.0
    if parameters.w0 == -1 and parameters.wa == 0:
        if is_flat:
            return cosmology.FlatLambdaCDM
        else:
            return cosmology.LambdaCDM
    if parameters.w0 != -1 and parameters.wa == 0:
        if is_flat:
            return cosmology.FlatwCDM
        else:
            return cosmology.wCDM
    if parameters.wa != 0:
        if is_flat:
            return cosmology.Flatw0waCDM
        else:
            return cosmology.w0waCDM

    raise ValueError("Could not determine cosmology type.")



from opencosmo.parameters import hacc


def get_origin_parameters(origin: str):
    if origin == "HACC":
        return hacc.ORIGIN_PARAMETERS
    else:
        raise ValueError(f"Unknown dataset origin {origin}")



from __future__ import annotations

from typing import TYPE_CHECKING, TypeGuard

import h5py
import numpy as np

from opencosmo.index import simple
from opencosmo.index.get import get_data_chunked

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from opencosmo.index.protocols import DataIndex


class ChunkedIndex:
    def __init__(self, starts: np.ndarray, sizes: np.ndarray) -> None:
        """
        Assumptions:
            - self.__starts is sorted
            - chunks are non-overlapping
        """
        self.__starts = starts
        self.__sizes = sizes

    @property
    def starts(self):
        return self.__starts

    @property
    def sizes(self):
        return self.__sizes

    def range(self) -> tuple[int, int]:
        """
        Get the range of the index.
        """
        if len(self) == 0:
            return 0, 0
        return self.__starts[0], self.__starts[-1] + self.__sizes[-1]

    def is_single_chunk(self):
        return len(self.__starts) == 1

    def into_array(self) -> NDArray[np.int_]:
        """
        Convert the ChunkedIndex to a SimpleIndex.
        """
        if len(self) == 0:
            return np.array([], dtype=int)
        idxs = np.concatenate(
            [
                np.arange(start, start + size)
                for start, size in zip(self.__starts, self.__sizes)
            ]
        )
        return np.unique(idxs)

    def into_mask(self) -> np.ndarray:
        ends = self.__starts + self.__sizes
        mask = np.zeros(np.max(ends), dtype=bool)
        for start, end in np.nditer([self.__starts, ends]):
            mask[start:end] = True
        return mask

    def intersection(self, other: DataIndex) -> DataIndex:
        if len(self) == 0 or len(other) == 0:
            return simple.SimpleIndex.empty()
        other_mask = other.into_mask()
        self_mask = self.into_mask()
        length = max(len(other_mask), len(self_mask))
        self_mask.resize(length)
        other_mask.resize(length)
        new_idx = np.where(self_mask & other_mask)[0]
        return simple.SimpleIndex(new_idx)

    def projection(self, other: DataIndex):
        if isinstance(other, simple.SimpleIndex):
            return self.__project_simple(other)
        elif isinstance(other, ChunkedIndex):
            return self.__project_chunked(other)

    def __project_simple(self, other: simple.SimpleIndex):
        self_simple = simple.SimpleIndex(self.into_array())
        return self_simple.projection(other)

    def __project_chunked(self, other: ChunkedIndex):
        self_ends = self.__starts + self.__sizes
        other_ends = other.__starts + other.__sizes
        out_of_range = (other_ends[:, np.newaxis] < self.__starts) | (
            other.__starts[:, np.newaxis] > self_ends
        )

        clipped_starts = np.clip(
            other.__starts[:, np.newaxis], a_min=self.__starts, a_max=None
        )
        clipped_ends = np.clip(other_ends[:, np.newaxis], a_min=None, a_max=self_ends)

        rs = np.cumsum(self.__sizes)
        rs = np.insert(rs, 0, 0)[:-1]

        starts = clipped_starts - self.__starts + rs
        ends = clipped_ends - self.__starts + rs
        sizes = ends - starts

        starts = starts.flatten()
        sizes = sizes.flatten()

        mask = sizes > 0
        return ChunkedIndex(starts[mask], sizes[mask])

    def concatenate(self, *others: DataIndex) -> DataIndex:
        if len(others) == 0 or all(len(o) == 0 for o in others):
            return self

        elif len(self) == 0:
            return others[0].concatenate(*others[1:])

        if all_are_chunked(others):
            new_starts = np.concatenate(
                [self.__starts] + [other.__starts for other in others]
            )
            new_sizes = np.concatenate(
                [self.__sizes] + [other.__sizes for other in others]
            )
            return ChunkedIndex(new_starts, new_sizes)

        else:
            indexes = np.concatenate([o.into_array() for o in others])
            indexes = np.concatenate((self.into_array(), indexes))
            return simple.SimpleIndex(indexes)

    @classmethod
    def from_size(cls, size: int) -> "ChunkedIndex":
        """
        Create a ChunkedIndex from a size.
        """
        if size <= 0:
            return ChunkedIndex.empty()
        # Create an array of chunk sizes

        starts = np.array([0])
        sizes = np.array([size])
        return ChunkedIndex(starts, sizes)

    @classmethod
    def single_chunk(cls, start: int, size: int) -> "ChunkedIndex":
        """
        Create a ChunkedIndex with a single chunk.
        """
        if start < 0:
            raise ValueError(f"Start must be non-negative, got {start}")

        if not size:
            return ChunkedIndex.empty()
        starts = np.array([start])
        sizes = np.array([size])
        return ChunkedIndex(starts, sizes)

    @classmethod
    def empty(cls):
        start = np.array([], dtype=int)
        size = np.array([], dtype=int)
        return ChunkedIndex(start, size)

    def set_data(self, data: np.ndarray, value: bool) -> np.ndarray:
        """
        Set the data at the index to the given value.
        """
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array")

        for start, size in zip(self.__starts, self.__sizes):
            data[start : start + size] = value
        return data

    def __len__(self) -> int:
        """
        Get the total size of the index.
        """
        return np.sum(self.__sizes)

    def mask(self, mask: np.ndarray) -> DataIndex:
        """
        Mask the index with a boolean mask.
        """
        if mask.shape != (len(self),):
            raise ValueError(
                f"Mask shape {mask.shape} does not match index size {len(self)}"
            )

        if mask.dtype != bool:
            raise ValueError(f"Mask dtype {mask.dtype} is not boolean")

        if not mask.any():
            return simple.SimpleIndex.empty()

        if mask.all():
            return self

        # Get the indices of the chunks that are masked
        idxs = np.concatenate(
            [
                np.arange(start, start + size)
                for start, size in zip(self.__starts, self.__sizes)
            ]
        )
        masked_idxs = idxs[mask]

        return simple.SimpleIndex(masked_idxs)

    def get_data(self, data: h5py.Dataset | np.ndarray) -> np.ndarray:
        """
        Get the data from the dataset using the index. We want to perform as few reads
        as possible. However, the chunks may not be continuous. This method sorts the
        chunks so it can read the data in the largest possible chunks, it then
        reshuffles the data to match the original order.

        For large numbers of chunks, this is much much faster than reading each chunk
        in the order they are stored in the index. I know because I tried. It sucked.
        """
        if not isinstance(data, (h5py.Dataset, np.ndarray)):
            raise ValueError("Data must be a h5py.Dataset")

        if len(self) == 0:
            return np.array([], dtype=data.dtype)
        if len(self.__starts) == 1:
            return data[self.__starts[0] : self.__starts[0] + self.__sizes[0]]

        return get_data_chunked(data, self.__starts, self.__sizes)

    def n_in_range(
        self, start: NDArray[np.int_], size: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        """
        Return the number of elements in this index that fall within
        a specified data range. Used to mask spatial index.


        As with numpy, this is the half-open range [start, end)
        """
        if len(start) != len(size):
            raise ValueError("Start and size arrays must have the same length")
        if np.any(size < 0):
            raise ValueError("Sizes must greater than or equal to zero")
        if len(self) == 0:
            return np.zeros_like(start)

        required_memory = len(start) * len(self.__starts) * 8
        MEMORY_LIMIT = 2_000_000_000
        if required_memory > MEMORY_LIMIT:
            chunk_size = MEMORY_LIMIT // (8 * len(self.__starts))
            n_chunks = np.ceil(len(start) / chunk_size)
        else:
            n_chunks = 1

        ends = start + size
        self_ends = self.__starts + self.__sizes

        rs = 0
        output = np.zeros_like(start)
        for arr in np.array_split(start, n_chunks):
            in_range_starts = np.clip(
                self.__starts[:, np.newaxis],
                a_min=arr,
                a_max=ends[rs : rs + len(arr)],
            )
            in_range_ends = np.clip(
                self_ends[:, np.newaxis],
                a_min=arr,
                a_max=ends[rs : rs + len(arr)],
            )
            output[rs : rs + len(arr)] = np.sum(in_range_ends - in_range_starts, axis=0)
            rs += len(arr)

        return output
        # example: row[0] in in_range_starts tells me what happens if I clip the 0th
        # element in self.__starts to each of the values in the start and end array
        # that have been provided in the call.

    def __getitem__(self, item: int) -> DataIndex:
        """
        Get an item from the index.
        """
        if item < 0 or item >= len(self):
            raise IndexError(
                f"Index {item} out of bounds for index of size {len(self)}"
            )
        sums = np.cumsum(self.__sizes)
        index = np.searchsorted(sums, item)
        start = self.__starts[index]
        offset = item - sums[index - 1] if index > 0 else item
        offset = int(offset)
        return simple.SimpleIndex(np.array([start + offset]))


def all_are_chunked(
    others: tuple[DataIndex, ...],
) -> TypeGuard[tuple[ChunkedIndex, ...]]:
    """
    Check if all elements in the tuple are instances of ChunkedIndex.
    """
    return all(isinstance(other, ChunkedIndex) for other in others) or not others



from __future__ import annotations

from typing import TYPE_CHECKING, Optional
from warnings import warn

from opencosmo.index import ChunkedIndex
from opencosmo.spatial.protocols import TreePartition

if TYPE_CHECKING:
    import h5py
    from mpi4py import MPI

    from opencosmo.spatial.tree import Tree


def partition(
    comm: MPI.Comm, length: int, counts: h5py.Group, tree: Optional[Tree]
) -> Optional[TreePartition]:
    """
    When opening with MPI, each rank recieves an equally-sized chunk of the
    spatial index. In principle this means the number of objects are similar
    between ranks.
    """
    if tree is not None:
        partitions = tree.partition(comm.Get_size(), counts)
        try:
            part = partitions[comm.Get_rank()]
        except IndexError:
            warn(
                "This MPI Rank recieved no data. "
                "The tree doesn't have enough subdivisions to serve every rank!"
            )
            part = None
        return part

    nranks = comm.Get_size()
    rank = comm.Get_rank()
    if rank == nranks - 1:
        start = rank * (length // nranks)
        size = length - start
        index = ChunkedIndex.single_chunk(start, size)

    else:
        start = rank * (length // nranks)
        end = (rank + 1) * (length // nranks)
        size = end - start

        index = ChunkedIndex.single_chunk(start, size)

    return TreePartition(index, None, None)



from __future__ import annotations

from collections import defaultdict
from itertools import chain
from typing import TYPE_CHECKING, Any, Iterable, Optional

import h5py
import hdf5plugin  # type: ignore

import opencosmo.io.writers as iow
from opencosmo.index import ChunkedIndex

if TYPE_CHECKING:
    import numpy as np
    from mpi4py import MPI
    from numpy.typing import DTypeLike, NDArray

    import opencosmo.io.protocols as iop
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.index import DataIndex

ColumnShape = tuple[int, ...]

"""
Schemas represent the hierarchy of groups and datasets in a given hdf file.

Schemas may have child schemas, and must be able to allocate themselves
in an open HDF5 file. Allocation may look as simple as creating a particular
group and passing that group to child allocators. In the case of a ColumnSchema,
actual creation of hdf5 datasets is performed, complete with its length, dtype
and compressin.

Any validaty tests across multiple children must be performed by the parent schema.
For example, an OpenCosmoDataset must have columns with compatible shapes. It is the
responsibility of the DatasetSchema to enforce this.


Note that "Dataset" is used in the opencosmo sense here so:

hdf5.Dataset == opencosmo Column
hdf5.Group == opencosmo Dataset (sometimes)

Schemas also must define an into_writer method that creates the object that will
actually put data into the file. As a result, certain schemas must hold references
to the underlying data they reference.

See opencosmo.io.protocols for required methods.
"""

COMPRESSION = hdf5plugin.Blosc2()


class FileSchema:
    """
    Always the top level of schema. Has very few responsibilites
    besides holding children.
    """

    def __init__(self):
        self.children = {}

    def add_child(self, child: iop.DataSchema, name: str):
        if name in self.children:
            raise ValueError(f"File schema already has child named {name}")
        match child:
            case (
                SimCollectionSchema()
                | StructCollectionSchema()
                | DatasetSchema()
                | LightconeSchema()
                | EmptyColumnSchema()
            ):
                self.children[name] = child
            case _:
                raise ValueError(
                    f"File schema cannot have children of type {type(child)}"
                )

    def allocate(self, group: h5py.File | h5py.Group):
        self.verify()
        if not isinstance(group, h5py.File):
            raise ValueError(
                "File Schema allocation must be done at the top level of a h5py file!"
            )
        if len(self.children) == 1:
            ds = next(iter(self.children.values()))
            return ds.allocate(group)

        for ds_name, ds in self.children.items():
            ds_group = group.require_group(ds_name)
            ds.allocate(ds_group)

    def verify(self):
        if len(self.children) == 0:
            raise ValueError("This file schema has no children!")
        for child in self.children.values():
            child.verify()

    def into_writer(self, comm: Optional["MPI.Comm"] = None):
        children = {
            name: schema.into_writer(comm) for name, schema in self.children.items()
        }
        return iow.FileWriter(children)


class SimCollectionSchema:
    """
    A schema for simulation collections. Like FileSchema, it has very few
    responsibilites outside of holding children and making sure they are
    valid types.
    """

    def __init__(self):
        self.children = {}

    def add_child(self, child: iop.DataSchema, name: str):
        if name in self.children:
            raise ValueError(f"SimCollection already has a dataset with name {name}")
        match child:
            case StructCollectionSchema() | DatasetSchema():
                self.children[name] = child
            case _:
                raise ValueError(
                    f"Sim collection cannot take children of type {type(child)}"
                )

    def allocate(self, group: h5py.File | h5py.Group):
        if not isinstance(group, h5py.File):
            raise ValueError(
                "File Schema allocation must be done at the top level of a h5py file!"
            )
        for ds_name, ds in self.children.items():
            ds_group = group.require_group(ds_name)
            ds.allocate(ds_group)

    def verify(self):
        zero_length = set()
        for name, child in self.children.items():
            try:
                child.verify()
            except ZeroLengthError:
                zero_length.add(name)
        if len(zero_length) == len(self.children):
            raise ZeroLengthError

    def into_writer(self, comm: Optional["MPI.Comm"] = None):
        children = {
            name: child.into_writer(comm) for name, child in self.children.items()
        }
        return iow.CollectionWriter(children)


class LightconeSchema:
    def __init__(self):
        self.children: dict[str, DatasetSchema] = {}

    def verify(self):
        zero_length = set()
        for name, child in self.children.items():
            try:
                child.verify()
            except ZeroLengthError:
                zero_length.add(name)
        if len(zero_length) == len(self.children):
            raise ZeroLengthError

    def add_child(self, child: iop.DataSchema, name: str):
        if name in self.children:
            raise ValueError(f"LightconeSchema already has child with name {name}")
        match child:
            case DatasetSchema():
                self.children[name] = child
            case _:
                raise ValueError(
                    f"LightconeSchema cannot take children of type {type(child)}"
                )

    def allocate(self, group: h5py.Group):
        if len(self.children) == 1:
            next(iter(self.children.values())).allocate(group)
            return
        for ds_name, ds in self.children.items():
            ds_group = group.require_group(ds_name)
            ds.allocate(ds_group)

    def into_writer(self, comm: Optional["MPI.Comm"] = None):
        dataset_writers = {
            key: val.into_writer(comm) for key, val in self.children.items()
        }
        return iow.CollectionWriter(dataset_writers)


class StructCollectionSchema:
    """
    Schema for structure collections. Actual linking data is held by datasets,
    so the main responsibility of this schema is to ensure that at least one
    link exists across all its datasets.

    Structure collections always have a header, because they always
    contain several datasets from the same simulation.
    """

    def __init__(self):
        self.children: dict[str, DatasetSchema | StructCollectionSchema] = defaultdict(
            dict
        )

    def verify(self):
        if len(self.children) < 2:
            raise ValueError("StructCollection must have at least two children!")

        found_links = False
        zero_length = set()
        for name, child in self.children.items():
            try:
                child.verify()
            except ZeroLengthError:
                zero_length.add(name)
                continue
            linked_columns = filter(lambda col: "data_linked" in col, child.children)
            if list(linked_columns):
                found_links = True

        if not found_links:
            raise ValueError("StructCollection must get at least one link!")
        elif len(zero_length) == len(self.children):
            raise ZeroLengthError

    def add_child(self, child: iop.DataSchema, name: str):
        if name in self.children:
            raise ValueError(
                f"StructCollectionSchema already has child with name {name}"
            )
        match child:
            case DatasetSchema():
                self.children[name] = child
            case StructCollectionSchema():
                for key, grandchild in child.children.items():
                    if key in self.children:
                        self.children[f"{name}_{key}"] = grandchild
                    else:
                        self.children[key] = grandchild
            case _:
                raise ValueError(
                    f"StructCollectionSchema cannot take children of type {type(child)}"
                )

    def allocate(self, group: h5py.Group):
        if len(self.children) == 1:
            raise ValueError("A dataset collection cannot have only one member!")
        for ds_name, ds in self.children.items():
            ds_group = group.require_group(ds_name)
            ds.allocate(ds_group)

    def into_writer(self, comm: Optional["MPI.Comm"] = None):
        dataset_writers = {
            key: val.into_writer(comm) for key, val in self.children.items()
        }
        return iow.CollectionWriter(dataset_writers)


class ZeroLengthError(Exception):
    pass


class DatasetSchema:
    """
    A schema for an opencosmo.dataset. This schema may or may not have a header.
    For example, a StructureCollection has a single collection for all datasets,
    so it does not need to be handled by the datasets it holds.
    """

    def __init__(
        self,
        header: Optional[OpenCosmoHeader] = None,
    ):
        self.columns: dict[str, dict[str, ColumnSchema]] = defaultdict(dict)
        self.header = header

    @property
    def children(self):
        return self.columns

    @classmethod
    def make_schema(
        cls,
        sources: dict[str, h5py.Group],
        columns: Iterable[str],
        index: DataIndex,
        cache: dict[str, np.ndarray],
        header: Optional[OpenCosmoHeader] = None,
    ):
        schema = DatasetSchema(header)
        columns = set(columns)
        for colname in columns:
            colname_parts = colname.split("/", maxsplit=1)
            if (
                colname_parts[0] not in sources
                or colname_parts[1] not in sources[colname_parts[0]].keys()
            ):
                raise ValueError("Found a column with no source!")

            # Still need this to get metadata if the column is cached
            colsource = sources[colname_parts[0]][colname_parts[1]]

            if colname in cache:
                coldata = cache[colname]
                column_schema = ColumnSchema(
                    colname_parts[1],
                    index.from_size(len(coldata)),
                    coldata,
                    colsource.attrs,
                )
            else:
                column_schema = ColumnSchema(
                    colname_parts[1], index, colsource, colsource.attrs
                )
            schema.add_child(column_schema, colname)
        return schema

    def add_child(self, child: iop.DataSchema, child_id: str):
        if not isinstance(child, ColumnSchema):
            raise ValueError("Dataset schemas can only take columns as childre")
        colparts = child_id.split("/", maxsplit=1)

        if colparts[1] in self.columns[colparts[0]]:
            raise ValueError(f"DatasetSchema already has a child named {child_id}")
        self.columns[colparts[0]][colparts[1]] = child

    def verify(self):
        if len(self.columns) == 0:
            raise ValueError("Datasets must have at least one column")

        column_lengths = set()
        for group, columns in self.columns.items():
            sizes = set(len(col) for col in columns.values())

        if len(sizes) == 1 and column_lengths.pop() == 0:
            raise ZeroLengthError()

        for group in self.columns.values():
            for column in group.values():
                column.verify()

    def allocate(self, group: h5py.File | h5py.Group):
        for groupname, columns in self.columns.items():
            data_group = group.require_group(groupname)
            for colname, column in columns.items():
                column.allocate(data_group)
        if self.header is not None:
            self.header.write(group)

    def into_writer(self, comm: Optional["MPI.Comm"] = None):
        colnames = list(self.columns.keys())
        colnames.sort()

        writers = {}
        for group, columns in self.columns.items():
            writers[group] = {
                name: cs.into_writer(comm) for name, cs in columns.items()
            }
        return iow.DatasetWriter(writers, comm)


class ColumnSchema:
    """
    This is where the magic actually happens. The ColumnSchema actually allocates
    space in the file holds a reference to the data that will eventually be written.
    It is also used eternally by the link schemas.
    """

    def __init__(
        self,
        name: str,
        index: DataIndex,
        source: h5py.Dataset | NDArray,
        attrs: dict[str, Any],
        total_length: Optional[int] = None,
    ):
        self.name = name
        self.index = index
        self.source = source
        self.attrs = attrs
        self.offset = 0
        if total_length is None:
            total_length = len(index)
        self.total_length = total_length

    def __len__(self):
        return len(self.index)

    def concatenate(self, *others: "ColumnSchema"):
        for other in others:
            if other.name != self.name:
                raise ValueError("Tried to combine columns with different names")
            if self.source.shape[1:] != other.source.shape[1:]:
                raise ValueError("Tried to combine columns with incompatible shapes")

        new_index = self.index.concatenate(*[o.index for o in others])
        return ColumnSchema(self.name, new_index, self.source, self.attrs)

    def add_child(self, *args, **kwargs):
        raise TypeError("Columns do not take children!")

    insert = add_child

    def set_offset(self, offset: int):
        self.offset = offset

    def verify(self):
        return True

    def allocate(self, group: h5py.Group):
        shape = (self.total_length,) + self.source.shape[1:]
        group.require_dataset(self.name, shape, self.source.dtype)
        for name, attr in self.attrs.items():
            group[self.name].attrs[name] = attr

    def into_writer(self, comm: Optional["MPI.Comm"] = None):
        return iow.ColumnWriter(self.name, self.index, self.source, self.offset)


class EmptyColumnSchema:
    """
    Represents a column that contains no data. The ONLY time this should be used
    is if we are writing in an MPI context and one or more of the ranks
    does not have data for a given column. However it must still know the column
    exists because structural work must be performed by all ranks in
    parallel HDF5.

    Its associated writer is effectively a no-op.

    """

    def __init__(
        self,
        name: str,
        attrs: dict[str, Any],
        dtype: DTypeLike,
        shape: tuple[int, ...],
    ):
        self.name = name
        self.attrs = attrs
        self.dtype = dtype
        self.shape = shape

    def __len__(self):
        return self.shape[0]

    def add_child(self, *args, **kwargs):
        raise TypeError("Columns do not take children!")

    insert = add_child

    def set_offset(self, offset: int):
        return

    def verify(self):
        return True

    def allocate(self, group: h5py.Group):
        group.create_dataset(self.name, self.shape, self.dtype)
        for name, attr in self.attrs.items():
            group[self.name].attrs[name] = attr

    def into_writer(self, comm: Optional["MPI.Comm"] = None):
        return iow.EmptyColumnWriter(self.name)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import astropy.units as u  # type: ignore
import numpy as np
from astropy.coordinates import SkyCoord  # type: ignore

from opencosmo.spatial.models import (
    BoxRegionModel,
    ConeRegionModel,
    HealPixRegionModel,
)
from opencosmo.spatial.region import (
    BoxRegion,
    ConeRegion,
    HealPixRegion,
)

if TYPE_CHECKING:
    from pydantic import BaseModel

    from opencosmo.spatial.region import (
        BoxSize,
        Point2d,
        Point3d,
    )


def from_model(model: BaseModel):
    match model:
        case ConeRegionModel():
            return make_cone(model.center, model.radius)
        case BoxRegionModel():
            return make_box(model.p1, model.p2)
        case HealPixRegionModel():
            return HealPixRegion(np.array(model.pixels), model.nside)
        case _:
            raise ValueError(f"Invalid region model type {type(model)}")


def make_box(p1: Point3d, p2: Point3d):
    """
    Create a 3-Dimensional box region of arbitrary size.
    The quantities of the box region are unitless, but will be converted
    to the unit convention of any dataset they interact with.

    Parameters
    ----------
    p1: (float, float, float)
        3D Point definining one corner of the box
    p1: (float, float, float)
        3D Point definining the other corner of the box

    Returns
    -------
    region: :py:class:`opencosmo.spatial.BoxRegion`
        The constructed region


    Raises
    ------
    ValueError
        If the region has zero length in any dimension
    """
    if len(p1) != 3 or len(p2) != 3:
        raise ValueError("Expected two 3-dimensional points")
    bl = tuple(min(p1d, p2d) for p1d, p2d in zip(p1, p2))
    tr = tuple(max(p1d, p2d) for p1d, p2d in zip(p1, p2))

    width = tuple(trd - bld for bld, trd in zip(bl, tr))
    center = tuple(bld + wd / 2.0 for bld, wd in zip(bl, width))
    if any(w == 0 for w in width):
        raise ValueError("At least one dimension of this box has zero length!")

    if isinstance(width, float) or isinstance(width, int):
        width = (width, width, width)

    halfwidth = cast("BoxSize", tuple(float(w / 2) for w in width))
    center = cast("Point3d", center)

    return BoxRegion(center, halfwidth)


def make_cone(center: Point2d | SkyCoord, radius: float | u.Quantity):
    """
    Construct a cone region used for querying lightcones. A cone
    region is defined by a location on the sky and an angular size,
    which is used as a radius.

    Parameters
    ----------
    center: astropy.coordinates.SkyCord | tuple[float, float]
        The center of the cone region. If a unitless tuple is passed,
        the values are assumed to be in degrees.

    radius: astropy.units.Quantity | float
        The radius of the region. If a unitless value is passed,
        it is assumed to be in degrees.

    Returns
    -------
    region: :py:class:`opencosmo.spatial.ConeRegion`
        The constructed region
    """
    coord: SkyCoord
    match center:
        case SkyCoord():
            coord = center
        case (float(ra) | int(ra), float(dec) | int(dec)):
            coord = SkyCoord(ra * u.deg, dec * u.deg)
        case (u.Quantity(), u.Quantity()):
            coord = SkyCoord(*center)
        case _:
            raise ValueError("Invalid center for Cone region")
    if isinstance(radius, (float, int)):
        radius = radius * u.deg
    return ConeRegion(coord, radius)



# ruff: noqa
from importlib import import_module

known_yt_tools = [
    "create_yt_dataset",
    "ProjectionPlot",
    "SlicePlot",
    "ParticleProjectionPlot",
    "ProfilePlot",
    "PhasePlot",
    "visualize_halo",
    "halo_projection_array",
]


"""
Right now, we have only have two analysis modules so we can handle them directly. In the 
future we will need to implement a more robust system that handles things automatically.
"""


def __getattr__(name):
    if name in known_yt_tools:
        yt_viz = import_module(".yt_viz", package="opencosmo.analysis")
        yt_utils = import_module(".yt_utils", package="opencosmo.analysis")
        try:
            return getattr(yt_viz, name)
        except AttributeError:
            pass
        try:
            return getattr(yt_utils, name)
        except AttributError:
            pass

        # except ImportError as ie:
        #    raise ImportError(
        #        "You tried to import one of the OpenCosmo YT tools, but your python "
        #        "environment does not have the necessary dependencies. You can do install "
        #        "them with `pip install opencosmo[analysis]`\n"
        #        f"{ie}"
        #    )
    raise ImportError(f"Cannot import name '{name}' from opencosmo.analysis")



from typing import Any, Sequence

import astropy.units as u
import numpy as np

"""
General helper routines for evaluating expressions on datasets and collections
"""


def insert(
    storage: dict[str, np.ndarray], index: int, values_to_insert: dict[str, Any]
):
    if isinstance(values_to_insert, dict):
        for name, value in values_to_insert.items():
            storage[name][index] = value
        return storage

    name = next(iter(storage.keys()))
    storage[name][index] = values_to_insert


def make_output_from_first_values(first_values: dict, n_rows: int):
    storage = {}
    for name, value in first_values.items():
        shape: tuple[int, ...] = (n_rows,)
        dtype = type(value)
        if isinstance(value, np.ndarray):
            shape = shape + value.shape
            dtype = value.dtype
        storage[name] = np.zeros(shape, dtype=dtype)
    for name, value in first_values.items():
        if isinstance(value, u.Quantity):
            storage[name] = storage[name] * value.unit
        storage[name][0] = value
    return storage


def prepare_kwargs(
    n_rows: int, evaluator_kwargs: dict[str, Any]
) -> tuple[dict[str, Any], dict[str, Sequence]]:
    kwargs = {}
    array_kwargs = {}
    for name, kwarg in evaluator_kwargs.items():
        try:
            length = len(kwarg)
            if length == n_rows:
                array_kwargs[name] = kwarg
                continue
            kwargs[name] = kwarg

        except TypeError:
            kwargs[name] = kwarg

    return kwargs, array_kwargs



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, TypeGuard

import h5py
import numpy as np

from opencosmo.index import chunked
from opencosmo.index.get import get_data_simple

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from opencosmo.index.protocols import DataIndex


class SimpleIndex:
    """
    An index of integers.
    """

    def __init__(self, index: NDArray[np.int_]) -> None:
        self.__index = np.sort(index)

    @classmethod
    def from_size(cls, size: int) -> "SimpleIndex":
        return SimpleIndex(np.arange(size))

    @classmethod
    def empty(cls):
        return SimpleIndex(np.array([], dtype=int))

    def __len__(self) -> int:
        return len(self.__index)

    def into_array(self, copy: bool = False) -> NDArray[np.int_]:
        if copy:
            return deepcopy(self.__index)
        return self.__index

    def range(self) -> tuple[int, int]:
        """
        Guranteed to be sorted
        """
        if len(self) == 0:
            return 0, 0
        return self.__index[0], self.__index[-1] + 1

    def into_mask(self):
        mask = np.zeros(self.__index[-1] + 1, dtype=bool)
        mask[self.__index] = True
        return mask

    def concatenate(self, *others: DataIndex) -> DataIndex:
        if len(others) == 0:
            return self

        indexes = np.concatenate([o.into_array() for o in others])
        indexes = np.concatenate((self.into_array(), indexes))
        return SimpleIndex(indexes)

    def n_in_range(
        self, start: NDArray[np.int_], size: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        if len(start) != len(size):
            raise ValueError("Start and size arrays must have the same length")
        if np.any(size < 0):
            raise ValueError("Sizes must greater than or equal to zero")
        if len(self) == 0:
            return np.zeros_like(start)

        ends = start + size
        start_idxs = np.searchsorted(self.__index, start, "left")
        end_idxs = np.searchsorted(self.__index, ends, "left")
        return end_idxs - start_idxs

    def set_data(self, data: np.ndarray, value: bool) -> np.ndarray:
        """
        Set the data at the index to the given value.
        """
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array")

        data[self.__index] = value
        return data

    def intersection(self, other: DataIndex) -> DataIndex:
        if len(self) == 0 or len(other) == 0:
            return SimpleIndex.empty()
        other_mask = other.into_mask()
        self_mask = self.into_mask()
        length = max(len(other_mask), len(self_mask))
        self_mask.resize(length)
        other_mask.resize(length)
        new_idx = np.where(self_mask & other_mask)[0]
        return SimpleIndex(new_idx)

    def projection(self, other: DataIndex):
        if isinstance(other, chunked.ChunkedIndex):
            other_simple = SimpleIndex(other.into_array())
            return self.projection(other_simple)

        isin = np.isin(self.into_array(), other.into_array())
        return SimpleIndex(np.where(isin)[0])

    def mask(self, mask: np.ndarray) -> DataIndex:
        if mask.shape != self.__index.shape:
            raise np.exceptions.AxisError(
                f"Mask shape {mask.shape} does not match index size {len(self)}"
            )

        if mask.dtype != bool:
            raise TypeError(f"Mask dtype {mask.dtype} is not boolean")

        if not mask.any():
            return SimpleIndex.empty()

        if mask.all():
            return self

        return SimpleIndex(self.__index[mask])

    def get_data(self, data: h5py.Dataset) -> np.ndarray:
        """
        Get the data from the dataset using the index.
        """
        if not isinstance(data, (h5py.Dataset, np.ndarray)):
            raise ValueError("Data must be a h5py.Dataset")
        if len(self) == 0:
            return np.array([], dtype=data.dtype)

        return get_data_simple(data, self.into_array())

    def __getitem__(self, item: int) -> DataIndex:
        """
        Get an item from the index.
        """
        if item < 0 or item >= len(self):
            raise IndexError(
                f"Index {item} out of bounds for index of size {len(self)}"
            )
        val = self.__index[item]
        return SimpleIndex(np.array([val]))


def all_are_simple(others: tuple[DataIndex, ...]) -> TypeGuard[tuple[SimpleIndex, ...]]:
    """
    Check if all elements in the tuple are instances of SimpleIndex.
    """
    return all(isinstance(other, SimpleIndex) for other in others) or not others



from __future__ import annotations

from typing import TYPE_CHECKING, Type, Union, get_origin

import h5py
import numpy as np
from pydantic import ValidationError

if TYPE_CHECKING:
    from pydantic import BaseModel


def read_header_attributes(
    file: h5py.File | h5py.Group,
    header_path: str,
    parameter_model: Type[BaseModel],
    **kwargs,
):
    header = file["header"]
    try:
        header_data = header[header_path].attrs
    except KeyError:
        return parameter_model()  # Defaults are possible
    try:
        parameters = parameter_model(**header_data, **kwargs)
    except ValidationError as e:
        msg = (
            "\nHeader attributes do not match the expected format for "
            f"{parameter_model.__name__}. "
            "Are you sure this is an OpenCosmo file?\n"
        )
        raise ValidationError.from_exception_data(msg, e.errors())  # type: ignore
    return parameters


def write_header_attributes(file: h5py.File, header_path: str, parameters: BaseModel):
    group = file.require_group(f"header/{header_path}")
    pars = parameters.model_dump(by_alias=True)
    for key, value in pars.items():
        if value is None:
            group.attrs[key] = ""
        else:
            group.attrs[key] = value

    return None


def get_empty_from_optional(type_: Type):
    origin = get_origin(type_)
    if origin is Union:
        np_equivalent = np.dtype(type_.__args__[0])
        return h5py.Empty(np_equivalent)
